/* $Id: cbq_config.c,v 1.3 1998/04/03 07:34:23 kjc Exp kjc $ */
/* derived from rtap of rsvp and rsvp_config of sun and then merged */
#pragma ident	"@(#)rsvp_config.c	1.16	97/01/20 SMI"
/* 
 * Copyright (c) Sun Microsystems, Inc.  1994. All rights reserved. 
 *
 * License is granted to copy, to use, and to make and to use derivative 
 * works for research and evaluation purposes, provided that Sun Microsystems is
 * acknowledged in all documentation pertaining to any such copy or derivative
 * work. Sun Microsystems grants no other licenses expressed or implied. The
 * Sun Microsystems  trade name should not be used in any advertising without
 * its written permission.
 *
 * SUN MICROSYSTEMS MERCHANTABILITY OF THIS SOFTWARE OR THE SUITABILITY OF
 * THIS SOFTWARE FOR ANY PARTICULAR PURPOSE.  The software is provided "as is"
 * without express or implied warranty of any kind.
 *
 * These notices must be retained in any copies of any part of this software.
 */
/*******************************************************************

  Copyright (c) 1996 by the University of Southern California
  All rights reserved.

  Permission to use, copy, modify, and distribute this software and its
  documentation in source and binary forms for any purpose and without
  fee is hereby granted, provided that both the above copyright notice
  and this permission notice appear in all copies. and that any
  documentation, advertising materials, and other materials related to
  such distribution and use acknowledge that the software was developed
  in part by the University of Southern California, Information
  Sciences Institute.  The name of the University may not be used to
  endorse or promote products derived from this software without
  specific prior written permission.

  THE UNIVERSITY OF SOUTHERN CALIFORNIA makes no representations about
  the suitability of this software for any purpose.  THIS SOFTWARE IS
  PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.

  Other copyrights might apply to parts of this software and are so
  noted when applicable.

********************************************************************/

#include <stddef.h>

#ifdef CBQDAEMON
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <syslog.h>
#include <sys/socket.h>
#include <net/if.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#else
#include "rsvp_daemon.h"
#include "rapi_lib.h"		/* Define flowspec formats */
#include "rsvp_specs.h"		/* Flowspec descriptor format */
#endif

#ifdef CBQDAEMON
#include "rsvp_cbq.h"
#include "if_nameindex.h"
#else
#include "tc_cbq.h"
#endif

#ifdef INCLUDE_SYSTEM_H
#include "system.h"
int printf(), strcasecmp();
#endif

#define show_help(op)  printf(cmd_tab[op].cmd_help)

/*
 * Forward & External Declarations
 */
static int	pfxcmp(char *, char *);
static int	next_word(char **, char *);

static int	cbq_do_cmd(int, char *);
static int	get_ifname(char **, char **);
static int	get_addr(char **, u_int *, u_int *);
static int	cbq_interface_cmd(char *);
static int	cbq_class_cmd(char *);
static int	cbq_filter_cmd(char *);
static int	cbq_ctl_cmd(char *);
static int	cbq_delete_cmd(char *);
#ifdef INET6
static int	cbq_filter6_cmd(char *);
static int	get_ip6addr(char **, struct in6_addr *, struct in6_addr *);
#endif

/*
 * Globals
 */
#define MAX_NFLWDS      64
#define MAX_T           64

int             TNO = 1;  	/* Current Thread number */

static char	if_names[MAX_T][IFNAMSIZ]; 
static int	line_no = 0;
static char	*configfile = "command input";
static struct if_nameindex *if_namelist = NULL;

#ifndef MAX
#define MAX(a,b) (((a)>(b))?(a):(b))
#endif
#ifndef MIN
#define MIN(a,b) (((a)<(b))?(a):(b))
#endif

enum op_codes {
        /* order must be same as entries cmd_tab[].cmd_op below!! */
        OP_HELP = 1, 	OP_QUIT, 
	OP_IFACE,	OP_CLASS,	OP_FILTER,
	OP_CBQ,		OP_DEL,		
#ifdef INET6
	OP_FILTER6,
#endif
        OP_NULL, 	OP_BUG
};

/*	Following table MUST match enum order of op_codes !
 */
struct cmds {
	char           *cmd_verb;
	int             cmd_op;
	char           *cmd_help;
}		cmd_tab[] = {

  	{ "?",		OP_HELP, 	"Commands are:\n" },
	{ "help",	OP_HELP, 	" help | ?\n" },
	{ "quit",	OP_QUIT, 	" quit\n" },
	{ "interface",	OP_IFACE,	" interface if_name [bandwidth bps] [sched_type]\n" },
	{ "class",	OP_CLASS,	" class cbq if_name class_name parent [borrow borrow_class] [admission ctlload|none] [maxburst count] [minburst count] [packetsize bytes] [priority pri] [pbandwidth percent] [red] [ecn]\n" },
	{ "filter",	OP_FILTER,	" filter if_name class_name dst [netmask #] dport src [netmask #] sport proto [tos # [tosmask #] [gpi #]\n" },
	{ "cbq",	OP_CBQ,		" cbq if_name {enable|disable|acc enable|acc disable}\n" },
	{ "delete",	OP_DEL,		" delete if_name class_name\n" },
#ifdef INET6
	{ "filter6",	OP_FILTER6,	" filter6 if_name class_name dst[/prefix] dport src[/prefix] sport proto [flowlabel #][tclass # [tclassmask #]][gpi #] \n" },
#endif
	{ "bug",	OP_BUG,		" bug (On/Off)\n" },
	{ "",		OP_NULL,	"" } /* MUST BE LAST IN CMD TABLE */
};

/*
 * Read the config file to learn about tunnel vifs and non-default phyint
 * parameters.
 */
int
cbq_config()
{
	FILE		*f;
	int		rc = 1;
	extern char	*solconfigfile;

	if (if_namelist != NULL)
		if_freenameindex(if_namelist);
	if_namelist = if_nameindex();
		
	f = fopen(solconfigfile, "r");
	if (f == NULL) {
		log(LOG_ERR, errno, "Can't open %s", solconfigfile, 0);
		return (-1);
	}
	line_no = 0;
	while (rc)
		rc = DoCommand(solconfigfile, f);

	(void) fclose(f);

	/* enable the interfaces */
	(void) cbq_enableall();

	return 0;
}

/*
 *	cbqd_cmds_init(): Initialization routine, called to start things off.
 */
void
cbq_cmds_init()
{
	int i;

	for (i = 0; i < MAX_T; i++)
		/* do initialize */
		if_names[i][0] = '\0';
}

/*
 *  Do_Command(): Top-level routine to read the next line from a given
 *	file and execute the command it contains.
 */
int
DoCommand(char *infile, FILE *infp)
{
	char	cmd_line[256], cmd_op[80];
	struct cmds *cmdp;
	char	*cp;
	int	rc;

	if (fgets(cmd_line, sizeof(cmd_line), infp) == NULL)
		/* EOF */
		return(0);
	line_no++;

	/* remove trailing NL */
	cp = cmd_line + strlen(cmd_line) - 1;
	if (*cp == '\n')
		*cp = '\0';
	else {
		printf("LINE > 255 CHARS: %s.\n", cmd_line);
		exit(1);
	}
	/*** printf("DoCommand: %s\n", cmd_line); ***/

	if (cmd_line[0] == '#') {	/* Comment, skip this line */
		return(1);
	}
	cp = cmd_line;
	if (!next_word(&cp, cmd_op))
		return(1);
	if (cmd_op[0] == 'T') {
		TNO = atoi(&cmd_op[1]);
		if (!next_word(&cp, cmd_op))
			return(1);
	}
	cmdp = cmd_tab;
	while ((cmdp->cmd_op != OP_NULL) && pfxcmp(cmd_op, cmdp->cmd_verb))
		cmdp++;

	if (cmdp->cmd_op == OP_NULL) {
		if (cmd_op[0])
			printf(" ?? %s\n", cmd_op);
		return(1);
	}
	rc = cbq_do_cmd(cmdp->cmd_op, cp);
	if (rc == 0) {
		if (infile) {
			/* error in the config file.  cleanup and exit. */
			log(LOG_ERR, 0, "Config failed. Exiting.\n");
			(void) cbq_destroyall();
			(void) fclose(infp);
			exit(1);
		}
		else {
			/* interactive mode */
			printf("error: usage :");
			show_help(cmdp->cmd_op);
		}
	}
	return(1);
}


/*
 * Prefix string comparison: Return 0 if s1 string is prefix of s2 string, 1
 * otherwise.
 */
static int 
pfxcmp(s1, s2)
	register char  *s1, *s2;
{
	while (*s1)
		if (*s1++ != *s2++)
			return (1);
	return (0);
}

/*
 * Skip leading blanks, then copy next word (delimited by blank or zero, but
 * no longer than 63 bytes) into buffer b, set scan pointer to following 
 * non-blank (or end of string), and return 1.  If there is no non-blank text,
 * set scan ptr to point to 0 byte and return 0.
 */
int 
next_word(char **cpp, char *b)
{
	char           *tp;
	int		L;

	*cpp += strspn(*cpp, " \t");
	if (**cpp == '\0' || **cpp == '\n' || **cpp == '#')
		return(0);

	tp = strpbrk(*cpp, " \t\n#");
	L = MIN((tp)?(tp-*cpp):strlen(*cpp), 63);
	strncpy(b, *cpp, L);
	*(b + L) = '\0';
	*cpp += L;
	*cpp += strspn(*cpp, " \t");
	return (1);
}

static int
cbq_do_cmd(int op, char *cmdbuf)
{
	int i, rval;
    
	switch (op) {
	case OP_HELP:
		for (i = 0; i < OP_NULL; i++)
			show_help(i);
		rval = 1;
		break;
	case OP_QUIT:
		cbq_destroyall();
		exit(0);
		break;
	case OP_IFACE:
		rval = cbq_interface_cmd(cmdbuf);
		break;
	case OP_CLASS:
		rval = cbq_class_cmd(cmdbuf);
		break;
	case OP_FILTER:
		rval = cbq_filter_cmd(cmdbuf);
		break;
	case OP_CBQ:
		rval = cbq_ctl_cmd(cmdbuf);
		break;
	case OP_DEL:
		rval = cbq_delete_cmd(cmdbuf);
		break;
#ifdef INET6
	case OP_FILTER6:
		rval = cbq_filter6_cmd(cmdbuf);
		break;
#endif
#ifdef CBQDAEMON
	case OP_BUG:
		if (debug & DEBUG_CBQ) {
			/* turn off verbose */
			l_debug = LOG_INFO;
			debug &= ~DEBUG_CBQ;
		}
		else {
			/* turn on verbose */
			l_debug = LOG_DEBUG;
			debug |= DEBUG_CBQ;
		}
		break;
#endif
	default:
		printf("cbq command %d not supported\n", op);
		rval = 0;
		break;
	}
	return(rval);
}

#define EQUAL(s1, s2)	(strcmp((char *)(s1), (char *)(s2)) == 0)

char *cur_ifname(void)
{
	return (if_names[TNO]);
}

static int
get_ifname(char **cpp, char **ifnamep)
{
	char w[128], *ocp;
	struct if_nameindex *ifnp;

	ocp = *cpp;
	if (next_word(&ocp, w) && if_namelist != NULL)
		for (ifnp = if_namelist; ifnp->if_name != NULL; ifnp++)
			if (strcmp(w, ifnp->if_name) == 0) {
				/* if_name found. advance the word pointer */
				*cpp = ocp; 
				strcpy(if_names[TNO], w);
				*ifnamep = if_names[TNO];
				return (1);
			}

	/* this is not interface name. use one in the context. */
	if (if_names[TNO][0] == 0)
		return (0);
	*ifnamep = if_names[TNO];
	return (1);
}

static int
get_addr(char **cpp, u_int *addr, u_int *mask)
{
	char w[128], *ocp;
	u_int tmp;
	
	*addr = 0;
	*mask = 0xffffffff;

	if (!next_word(cpp, w))
		return (0);

	if ((tmp = inet_addr((char *)w)) == INADDR_NONE)
		return (0);
	*addr = tmp;

	/* check if netmask option is present */
	ocp = *cpp;
	if (next_word(&ocp, w) && EQUAL(w, "netmask")) {
		if (!next_word(&ocp, w))
			return (0);
		
		if ((tmp = inet_addr((char *)w)) == INADDR_NONE)
			return (0);

		*mask = tmp;
		*cpp = ocp;	
		return (1);
	}
	/* no netmask option */
	return (1);
}

#define	DISCIPLINE_NONE		0
#define	DISCIPLINE_CBQ		1
#define DISCIPLINE_CSZ		2
#define DISCIPLINE_CTLLOAD	3

static int
cbq_interface_cmd(char *cmdbuf) 
{
	char	w[128], *cp = cmdbuf;
	char	*ifname;
	long   	bandwidth = 0;
	u_int	is_efficient = 0;
	int	is_sched = 1;
	int     discipline = DISCIPLINE_NONE;
	u_int	is_wrr = 1;	/* weighted round-robin is default */
    
#ifdef CBQDAEMON
	extern int Debug_mode;
	if (Debug_mode)
		is_sched = 0;
#endif
	if (!get_ifname(&cp, &ifname)) {
		log(LOG_ERR, 0, "missing interface name in %s, line %d",
		    configfile, line_no);
		return (0);
	}

	/*
	 * Look for "bandwidth" & scheduling discipline options.
	 */
	while (next_word(&cp, w)) {
		if (EQUAL(w, "bandwidth")) {
			next_word(&cp, w);
			bandwidth = atol(w);
		} else if (EQUAL(w, "efficient")) {
			is_efficient = 1;
		} else if (EQUAL(w, "cbq")) {
			discipline = DISCIPLINE_CBQ;
		} else if (EQUAL(w, "cbq-wrr")) {
			discipline = DISCIPLINE_CBQ;
			is_wrr = 1;
		} else if (EQUAL(w, "cbq-prr")) {
			discipline = DISCIPLINE_CBQ;
			is_wrr = 0;
		} else if (EQUAL(w, "admission_only")) {
			discipline = DISCIPLINE_CBQ;
			is_sched = 0;
		} else if (EQUAL(w, "nosched")) {
			is_sched = 0;
		} else if (EQUAL(w, "csz")) {
			log(LOG_ERR, 0,
			    "'csz' not yet supported - %s, line %d",
			    configfile, line_no);
			return (0);
		} else {
			log(LOG_ERR, 0,
			    "Unknown keyword '%s' in %s, line %d",
			    w, configfile, line_no);
			return (0);
		}
	}
	if (discipline == DISCIPLINE_CBQ) {
		/* Add this to list of CBQ interfaces */
		if (cbq_add_if(ifname, bandwidth, is_sched, is_wrr,
			       is_efficient)) {
			log(LOG_ERR, 0,
			    "can't init cbq on interface '%s'", ifname);
			return (0);
		}
	}
	else if (discipline == DISCIPLINE_NONE)
		log(LOG_WARNING, 0, "No discipline (e.g. cbq) specified!");
	else {
		log(LOG_ERR, 0, "unknown discipline!");
		return (0);
	}
	return (1);
}

static int
cbq_class_cmd(char *cmdbuf) 
{
	char	w[128], *cp = cmdbuf;
	char 	*ifname, class_name[128], parent_name[128], borrow_name[128];
	u_char  *clname = class_name;
	u_char  *parent = NULL;
	u_char  *borrow = NULL;
	u_int   pri = 1;
	int	flags = 0;
	u_int	pbandwidth = 100;
	u_int	maxdelay = 0;		/* 0 means default */
	u_int	maxburst = 16;
	u_int	minburst = 2;
	u_int	av_pkt_size = 0;  /* 0 means use if mtu as default */
	u_int	max_pkt_size = 0; /* 0 means use if mtu as default */
	cbq_tos_t	admission_type = CBQ_QOS_NONE;
	cbq_ol_t	overlimit_action = OL_DELAY;
	int	ret;

	/* get scheduling class */
	if (!next_word(&cp, w)) {
		log(LOG_ERR, 0, "missing scheduling discipline in %s, line %d",
		    configfile, line_no);
		return (0);
	}
	if (!EQUAL(w, "cbq")) {
		log(LOG_ERR, 0,
		    "unknown scheduling discipline '%s' in %s, line %d",
		    w, configfile, line_no);
		return (0);
	}

	/* get interface name */
	if (!get_ifname(&cp, &ifname)) {
		log(LOG_ERR, 0, "missing interface name in %s, line %d",
		    configfile, line_no);
		return (0);
	}

	/* get class name */
	if (!next_word(&cp, class_name)) {
		log(LOG_ERR, 0, "missing class name in %s, line %d",
		    configfile, line_no);
		return (0);
	}

	/* get parent name */
	if (!next_word(&cp, parent_name)) {
		log(LOG_ERR, 0, "missing parent class in %s, line %d",
		    configfile, line_no);
		return (0);
	}
	if (!EQUAL(parent_name, "null") && !EQUAL(parent_name, "NULL")) {
		parent = parent_name;
	}
	else {
		parent = NULL;
		flags |= CBQCLF_ROOTCLASS;
	}

	/* look for options. */
	while (next_word(&cp, w)) {
		if (EQUAL(w, "priority")) {
			next_word(&cp, w);
			pri = (u_int)atoi(w);
		} else if (EQUAL(w, "default")) {
			flags |= CBQCLF_DEFCLASS;
		} else if (EQUAL(w, "control")) {
			flags |= CBQCLF_CTLCLASS;
		} else if (EQUAL(w, "admission")) {
			next_word(&cp, w);
			if (EQUAL(w, "guaranteed"))
				admission_type = CBQ_QOS_GUARANTEED;
			else if (EQUAL(w, "predictive"))
				admission_type = CBQ_QOS_PREDICTIVE;
			else if (EQUAL(w, "cntlload"))
				admission_type = CBQ_QOS_CNTR_LOAD;
			else if (EQUAL(w, "cntldelay"))
				admission_type = CBQ_QOS_CNTR_DELAY;
			else if (EQUAL(w, "none"))
				admission_type = CBQ_QOS_NONE;
			else {
				log(LOG_ERR, 0,
				    "unknown admission type - %s, line %d",
				    w, line_no);
				return (0);
			}
		} else if (EQUAL(w, "maxdelay")) {
			next_word(&cp, w);
			maxdelay = (u_int)atoi(w);
		} else if (EQUAL(w, "borrow")) {
			next_word(&cp, borrow_name);
			borrow = borrow_name;
		} else if (EQUAL(w, "pbandwidth")) {
			next_word(&cp, w);
			pbandwidth = (u_int)atoi(w);
		} else if (EQUAL(w, "maxburst")) {
			next_word(&cp, w);
			maxburst = (u_int)atoi(w);
		} else if (EQUAL(w, "minburst")) {
			next_word(&cp, w);
			minburst = (u_int)atoi(w);
		} else if (EQUAL(w, "packetsize")) {
			next_word(&cp, w);
			av_pkt_size = (u_int)atoi(w);
		} else if (EQUAL(w, "maxpacketsize")) {
			next_word(&cp, w);
			max_pkt_size = (u_int)atoi(w);
		} else if (EQUAL(w, "red")) {
			flags |= CBQCLF_RED;
		} else if (EQUAL(w, "ecn")) {
			flags |= (CBQCLF_RED | CBQCLF_ECN);
		} else {
			log(LOG_ERR, 0,
			    "Unknown keyword '%s' in %s, line %d",
			    w, configfile, line_no);
			return (0);	
		}
	}
 
	ret = cbq_add_static_class(ifname, clname,
				   parent, borrow, pri, flags, pbandwidth,
				   maxdelay, maxburst, minburst,
				   av_pkt_size, max_pkt_size,
				   admission_type, overlimit_action);
    
	if (ret) {
		log(LOG_ERR, 0, "can't add class '%s' on interface '%s'",
		    clname, ifname);
		return (0);
	}
    
	return (1);
}

static int
cbq_filter_cmd(char *cmdbuf) 
{
	char 	w[128], *cp = cmdbuf;
	char 	*ifname, class_name[128];
	cbq_filt_t	sfilt;
	u_int	protocol;
	u_char	tos, tosmask;
	int	ret;

	memset(&sfilt, 0, sizeof(sfilt));
	sfilt.family = AF_INET;

	if (!get_ifname(&cp, &ifname)) {
		log(LOG_ERR, 0, "missing interface name in %s, line %d",
		    configfile, line_no);
		return (0);
	}

	if (!next_word(&cp, class_name)) {
		log(LOG_ERR, 0,
		    "missing class name in %s, line %d",
		    configfile, line_no);
		return (0);
	}

	/* get filter destination Address */
	if (!get_addr(&cp, &sfilt.daddr, &sfilt.dmask)) {
		log(LOG_ERR, 0,
		    "bad filter destination address in %s, line %d",
		    configfile, line_no);
		return (0);
	}

	/* get filter destination port */
	if (!next_word(&cp, w)) {
		log(LOG_ERR, 0,
		    "missing filter destination port in %s, line %d",
		    configfile, line_no);
		return (0);
	}
	if ((sscanf(w, "%hd", &sfilt.dport)) == 0) {
		log(LOG_ERR, 0, "bad filter destination port in %s, line %d",
		    configfile, line_no);
		return (0);
	}
	sfilt.dport = hton16(sfilt.dport);
    
	/* get filter source address */
	if (!get_addr(&cp, &sfilt.saddr, &sfilt.smask)) {
		log(LOG_ERR, 0, "bad filter source address in %s, line %d",
		    configfile, line_no);
		return (0);
	}

	/* get filter source port */
	if (!next_word(&cp, w)) {
		log(LOG_ERR, 0, "missing filter source port in %s, line %d",
		    configfile, line_no);
		return (0);
	}
	if ((sscanf(w, "%hd", &sfilt.sport)) == 0) {
		log(LOG_ERR, 0, "bad filter source port in %s, line %d",
		    configfile, line_no);
		return (0);
	}
	sfilt.sport = hton16(sfilt.sport);

	/* get filter protocol id */
	if (!next_word(&cp, w)) {
		log(LOG_ERR, 0, "missing filter protocol id in %s, line %d",
		    configfile, line_no);
		return (0);
	}
	if ((sscanf(w, "%d", &protocol)) == 0) {
		log(LOG_ERR, 0, "bad protocol in %s, line %d",
		    configfile, line_no);
		return (0);
	}
	sfilt.proto = protocol; 

	while (next_word(&cp, w)) {
		if (EQUAL(w, "tos")) {
			tos = 0;
			tosmask = 0xff;
			
			if (next_word(&cp, w)) {
				tos = (u_char)atoi(w);
				if (next_word(&cp, w)) {
					if (EQUAL(w, "tosmask")) {
						next_word(&cp, w);
						tosmask = (u_char)atoi(w);
					}
				}
			}
			sfilt.tos = tos; 
			sfilt.tosmask = tosmask; 
		}
		else if (EQUAL(w, "gpi")) {
			if (next_word(&cp, w)) {
				sfilt.gpi = atoi(w);
				sfilt.gpi = htonl(sfilt.gpi);
			}
		}
	}
	
	/*
	 * Add the filter.
	 */
	ret = cbq_add_static_filter(ifname, class_name, &sfilt);
	if (ret) {
		log(LOG_ERR, 0,
		    "can't add filter to class '%s' on interface '%s'",
		    class_name, ifname);
		return (0);
	}
	return (1);
}

#ifdef INET6
static int
cbq_filter6_cmd(char *cmdbuf)
{
	char 	w[128], *cp = cmdbuf;
	char 	*ifname, class_name[128];
	cbq_filt_t	sfilt;
	u_int	protocol;
	u_char	tos, tosmask;
	int	ret;

	memset(&sfilt, 0, sizeof(sfilt));
	sfilt.family = AF_INET6;

	if (!get_ifname(&cp, &ifname)) {
		log(LOG_ERR, 0, "missing interface name in %s, line %d",
		    configfile, line_no);
		return (0);
	}

	if (!next_word(&cp, class_name)) {
		log(LOG_ERR, 0, "missing class name in %s, line %d",
		    configfile, line_no);
		return (0);
	}

	/* get filter destination address */
	if (!get_ip6addr(&cp, &sfilt.daddr6, &sfilt.dmask6)) {
		log(LOG_ERR, 0, "bad destination address in %s, line %d",
		    configfile, line_no);
		return (0);
	}
	
	/* get filter destination port */
	if (!next_word(&cp, w)) {
		log(LOG_ERR, 0,
		    "missing filter destination port in %s, line %d",
		    configfile, line_no);
		return (0);
	}
	if ((sscanf(w, "%hd", &sfilt.dport6)) == 0) {
		log(LOG_ERR, 0, "bad filter destination port in %s, line %d",
		    configfile, line_no);
		return (0);
	}
	sfilt.dport6 = hton16(sfilt.dport6);
    
	/* get filter source address */
	if (!next_word(&cp, w)) {
		log(LOG_ERR, 0, "missing filter source address in %s, line %d",
		    configfile, line_no);
		return (0);
	}
	if (!get_ip6addr(&cp, &sfilt.saddr6, &sfilt.smask6)) {
		log(LOG_ERR, 0, "bad source address in %s, line %d",
		    configfile, line_no);
		return (0);
	}

	/* get filter source port */
	if (!next_word(&cp, w)) {
		log(LOG_ERR, 0, "missing filter source port in %s, line %d",
		    configfile, line_no);
		return (0);
	}
	if ((sscanf(w, "%hd", &sfilt.sport6)) == 0) {
		log(LOG_ERR, 0, "bad filter source port in %s, line %d",
		    configfile, line_no);
		return (0);
	}
	sfilt.sport6 = hton16(sfilt.sport6);

	/* get filter protocol id */
	if (!next_word(&cp, w)) {
		log(LOG_ERR, 0, "missing filter protocol id in %s, line %d",
		    configfile, line_no);
		return (0);
	}
	if ((sscanf(w, "%d", &protocol)) == 0) {
		log(LOG_ERR, 0, "bad protocol in %s, line %d",
		    configfile, line_no);
		return (0);
	}
	sfilt.proto = protocol; 

	while (next_word(&cp, w)) {
		if (EQUAL(w, "tclass")) {
			tos = 0;
			tosmask = 0xff;

			if (next_word(&cp, w)) {
				tos = (u_char)atoi(w);
				if (next_word(&cp, w)) {
					if (EQUAL(w, "tclassmask")) {
						next_word(&cp, w);
						tosmask = (u_char)atoi(w);
					}
				}
			}
			sfilt.tos = tos; 
			sfilt.tosmask = tosmask; 
		}
		else if (EQUAL(w, "gpi")) {
			if (next_word(&cp, w)) {
				sfilt.gpi6 = atoi(w);
				sfilt.gpi6 = htonl(sfilt.gpi6);
			}
		}
		else if (EQUAL(w, "flowlabel")) {
			if (next_word(&cp, w)) {
				sfilt.flowlabel6 = atoi(w) & 0x000fffff;
				sfilt.flowlabel6 = htonl(sfilt.flowlabel6);
			}
		}

	}

	/*
	 * Add the filter.
	 */
	ret = cbq_add_static_filter(ifname, class_name, &sfilt);
	if (ret) {
		log(LOG_ERR, 0,
		    "can't add filter to class '%s' on interface '%s'",
		    class_name, ifname);
		return (0);
	}
	return (1);
}

static int
get_ip6addr(char **cpp, struct in6_addr *addr, struct in6_addr *mask)
{
	char w[128], *prefix;
	u_char *cp;
	int len;

	memset(addr, 0, sizeof(struct in6_addr));
	memset(mask, 0, sizeof(struct in6_addr));

	if (!next_word(cpp, w))
		return (0);

	if ((prefix = strchr(w, '/')) != NULL) {
		/* address has prefix length */
		*prefix++ = '\0';
	}

	if (inet_pton(AF_INET6, w, addr) <= 0)
		return (0);

	/* convert address prefix length to address mask */
	if (prefix != NULL) {
		len = atoi(prefix);
		if ((len < 0) || (len > 128))
			return (0);
		for (cp = (u_char *)mask; len > 7; len -= 8)
			*cp++ = 0xff;
		if (len > 0)
			*cp = (0xff << (8 - len)) & 0xff;
	}

	return (1);
}

#endif /* INET6 */

static int
cbq_ctl_cmd(char *cmdbuf)
{
	char	w[128], *cp = cmdbuf;
	char	*ifname;
	int	cbq_state, acc_state;
	int	rval;
    
	if (!get_ifname(&cp, &ifname)) {
		printf("missing interface name in %s, line %d",
		       configfile, line_no);
		return (0);
	}

	if (!next_word(&cp, w)) {
		cbq_state = cbq_is_enabled(ifname);
		acc_state = cbq_is_acc_enabled(ifname);
		printf("cbq %s (accounting %s) on %s\n",
		       cbq_state ? "enabled" : "disabled",
		       acc_state ? "enabled" : "disabled", ifname);
		return (1);
	}
	
	if (EQUAL(w, "enable")) {
		rval = cbq_enable(ifname);
		printf("cbq %s on %s\n",
		       (rval == 0) ? "enabled" : "enable failed!", ifname);
	}
	else if (EQUAL(w, "disable")) {
		rval = cbq_disable(ifname);
		printf("cbq %s on %s\n",
		       (rval == 0) ? "disabled" : "disable failed!", ifname);
	}
	else if (EQUAL(w, "acc")) {
		if (!next_word(&cp, w))
			return (0);

		if (EQUAL(w, "enable")) {
			rval = cbq_acc_enable(ifname);
			printf("cbq accounting %s on %s\n",
			   (rval == 0) ? "enabled" : "enable failed!", ifname);
		}
		else if (EQUAL(w, "disable")) {
			rval = cbq_acc_disable(ifname);
			printf("cbq accounting %s on %s\n",
			 (rval == 0) ? "disabled" : "disable failed!", ifname);
		}
		else
			return (0);
	}
	else if (EQUAL(w, "reload")) {
		printf("reinitializing cbq...\n");
		cbq_destroyall();
		cbq_init();
	}
	else
		return (0);
	return (1);
}


static int
cbq_delete_cmd(char *cmdbuf)
{
	char	*cp = cmdbuf;
	char	*ifname, class_name[128];
	int	ret;
    
	if (!get_ifname(&cp, &ifname)) {
		printf("missing interface name in %s, line %d",
		       configfile, line_no);
		return (0);
	}

	if (!next_word(&cp, class_name)) {
		log(LOG_ERR, 0,
		    "missing class name in %s, line %d",
		    configfile, line_no);
		return (0);
	}

	ret = cbq_delete_static_class(ifname, class_name);
	if (ret) {
		log(LOG_ERR, 0,
		    "can't delete class '%s' on interface '%s'",
		    class_name, ifname);
		return (0);
	}

	return (1);
}

