/*
 * Copyright (C) 1997
 *	Sony Computer Science Laboratory Inc.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY SONY CSL AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL SONY CSL OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: ifspeed.c,v 0.2 1997/03/27 11:36:21 kjc Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <netdb.h>
#include <errno.h>
#include <varargs.h>
#include <sys/param.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/sockio.h>
#include <net/if.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#define MAX_LOOPCOUNT	64

int timeup = 0;
int sockfd = -1;
char buffer[8192];
struct sockaddr_in my_addr, dst_addr;
int nobufs = 0;
int packets = 0;


void usage()
{
    printf("usage: ifspeed ifname [-p dport] [-l loop_cnt] dest_host\n");
    exit(1);
}

void fatal_error(va_alist)
    va_dcl
{
    va_list ap;
    char *fmt;

    if (errno != 0)
	perror("fatal_error");
    else
	fprintf(stderr, "fatal_error: ");
    va_start(ap);
    fmt = va_arg(ap, char *);
    vfprintf(stderr, fmt, ap);
    va_end(ap);
    fprintf(stderr, "\n");
    exit(1);
}


void
alrm_handler(int sig)
{
    timeup = 1;
}

/* convert a string name to struct sockaddr_in */
int name2sockaddrin(char *name, int port, struct sockaddr_in *addrp)
{
    unsigned long inaddr;
    struct hostent *hep;

    bzero(addrp, sizeof(struct sockaddr_in));
    addrp->sin_family = PF_INET;
    if (name != NULL) {
	if ((inaddr = inet_addr(name)) != -1)
	    bcopy(&inaddr, &addrp->sin_addr, sizeof(inaddr));
	else if ((hep = gethostbyname(name)) != NULL)
	    bcopy(hep->h_addr, &addrp->sin_addr, hep->h_length);
	else
	    return (-1);
    }
    else
	addrp->sin_addr.s_addr = htonl(INADDR_ANY);
    addrp->sin_port = htons(port);
    return 0;
}

int
mainloop(int size)
{
    int cc, bytes = 0;
    
    while (!timeup) {
	if ((cc = sendto(sockfd, buffer, size, 0,
		   (struct sockaddr *)&dst_addr, sizeof(dst_addr))) == -1) {
	    if (errno == ENOBUFS)
		nobufs++;
	    else
		perror("sendto");
	}
	else {
	    bytes += cc;
	    packets++;
	}
    }
    return (bytes);
}


int
main(int argc, char **argv)
{
    char *dst_name = NULL;
    char *ifname = NULL;
    int size, bytes, msec, sec = 2;
    struct ifreq ifr;
    int dst_port = 9999;
    double ifspeed;
    struct timeval start_time, end_time;
    struct sockaddr_in *sin;
    int i, loop_count = 3;
    int results[MAX_LOOPCOUNT];

    if (argc < 3)
	usage();

    ifname = argv[1];

    argc -= 2;
    argv += 2;
    while(argc-- > 0) {
	if (argv[0][0] == '-') {
	    switch (argv[0][1]) {
	    case 'p':
		--argc;
		dst_port = atoi(*++argv);
		printf("dst port set to %d\n", dst_port);
		break;
	    case 'l':
		--argc;
		loop_count = atoi(*++argv);
		if (loop_count > MAX_LOOPCOUNT)
		    loop_count =  MAX_LOOPCOUNT;
		printf("loop count set to %d\n", loop_count);
		break;
	    default:
		usage();
		break;
	    }
	}
	else {
	    dst_name = *argv;
	}
	++argv;
    }

    if (ifname == NULL || dst_name == NULL)
	usage();
    
    (void)signal(SIGALRM, alrm_handler);
    
    if ((sockfd = socket(PF_INET, SOCK_DGRAM, 0)) < 0)
	fatal_error("can't open socket");

    /* get mtu of the interface */
    strncpy(ifr.ifr_name, ifname, sizeof(ifr.ifr_name));
    if (ioctl(sockfd, SIOCGIFMTU, (caddr_t)&ifr) < 0) {
	perror("ioctl SIOCGIFMTU");
	size = 512;
    }
    else
	size = ifr.ifr_mtu - 28;  /* IP header(20) + UDP header(8) */
    if (size > sizeof(buffer))
	size = sizeof(buffer);

    if (ioctl(sockfd, SIOCGIFADDR, (char *)&ifr) < 0) {
	fatal_error("ioctl SIOCGIFADDR");
    }

    bzero(&my_addr, sizeof(my_addr));
    my_addr.sin_family = PF_INET;
    sin = (struct sockaddr_in *)&ifr.ifr_addr;
    my_addr.sin_addr = sin->sin_addr;
    my_addr.sin_port = htons(0);  /* arbitrary port */

    if (name2sockaddrin(dst_name, dst_port, &dst_addr) < 0)
	fatal_error("can't get viewer's address!");

    if (bind(sockfd, (struct sockaddr *)&my_addr, sizeof(my_addr)) < 0)
	fatal_error("sender: can't bind");

    for (i=0; i<loop_count; i++) {
	printf("test %d/%d: start sending to %s:%d packetsize: %d\n",
	       i+1, loop_count, dst_name, dst_port, size);

	timeup = 0;
	packets = 0;
	nobufs = 0;
	
	alarm(sec);
	sec = sec * 2;  /* twice as long in the next loop */

	gettimeofday(&start_time, NULL);

	bytes = mainloop(size);

	gettimeofday(&end_time, NULL);

	if((end_time.tv_usec = end_time.tv_usec - start_time.tv_usec) < 0) {
	    end_time.tv_usec += 1000000;
	    end_time.tv_sec = end_time.tv_sec - start_time.tv_sec - 1;
	} else
	    end_time.tv_sec = end_time.tv_sec - start_time.tv_sec;

	msec = end_time.tv_sec * 1000 + end_time.tv_usec / 1000;

	ifspeed = (double)bytes * 8.0 / msec * 1000;

	printf("%d bytes sent in %.3f sec\n", bytes, (double)msec / 1000.0);
	printf("  (%d success, %d dropped)\n",
	       packets, nobufs);
	printf("\ninterface speed is %d bps\n\n", (int)ifspeed);
	results[i] = (int)ifspeed;

	sleep(3);
    }

    ifspeed = 0.0;
    for (i=0; i<loop_count; i++)
	ifspeed += (double)results[i];
    ifspeed = ifspeed / loop_count;

    printf("\nAverage Interface Speed is [%d] bps\n\n", (int)ifspeed);
    
    close(sockfd);
    exit(0);
}

