#pragma ident	"@(#)rsvp_cbq.c	1.42	97/08/06 SMI"
/*
 * @(#) $Id: rsvp_cbq.c,v 0.9 1998/04/03 07:34:23 kjc Exp kjc $
 */

/************************ rsvp_kern.c  *******************************
 *                                                                   *
 *  Interface between RSVP and traffic control modules in kernel.    *
 *     This module interfaces to CBQ, the Sun/UCL/LBL kernel.        *
 *                                                                   *
 *********************************************************************/

/****************************************************************************

            RSVPD -- ReSerVation Protocol Daemon

                Written by: Steven Berson (berson@isi.edu)
                USC Information Sciences Institute
                Marina del Rey, California

            Original Version: Shai Herzog, Nov. 1993.
	    Current Version:  Steven Berson, August 1994.


  Copyright (c) 1993 by the University of Southern California
  All rights reserved.

  Permission to use, copy, modify, and distribute this software and its
  documentation in source and binary forms for non-commercial purposes
  and without fee is hereby granted, provided that the above copyright
  notice appear in all copies and that both the copyright notice and
  this permission notice appear in supporting documentation. and that
  any documentation, advertising materials, and other materials related
  to such distribution and use acknowledge that the software was
  developed by the University of Southern California, Information
  Sciences Institute.  The name of the University may not be used to
  endorse or promote products derived from this software without
  specific prior written permission.

  THE UNIVERSITY OF SOUTHERN CALIFORNIA makes no representations about
  the suitability of this software for any purpose.  THIS SOFTWARE IS
  PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.

  Other copyrights might apply to parts of this software and are so
  noted when applicable.

********************************************************************/

/* 
 * Copyright (c) Sun Microsystems, Inc.  1996. All rights reserved. 
 *
 * License is granted to copy, to use, and to make and to use derivative 
 * works for research and evaluation purposes, provided that Sun Microsystems is
 * acknowledged in all documentation pertaining to any such copy or derivative
 * work. Sun Microsystems grants no other licenses expressed or implied. The
 * Sun Microsystems  trade name should not be used in any advertising without
 * its written permission.
 *
 * SUN MICROSYSTEMS MERCHANTABILITY OF THIS SOFTWARE OR THE SUITABILITY OF
 * THIS SOFTWARE FOR ANY PARTICULAR PURPOSE.  The software is provided "as is"
 * without express or implied warranty of any kind.
 *
 * These notices must be retained in any copies of any part of this software.
 */

#include <sys/param.h>
#include <stddef.h>

#ifdef CBQDAEMON
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <err.h>
#include <syslog.h>
#include <sys/socket.h>
#include <net/if.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#else
#include "rsvp_daemon.h"
#include "rapi_lib.h"		/* Define flowspec formats */
#include "rsvp_specs.h"		/* Flowspec descriptor format */
#include "rsvp_TCif.h"		/* Adaptation interface */

/*
 * tc interface was changed since rel4.2a2.  if LLDAL_RC_OK is defined,
 * use a new interface.
 */
#ifdef LLDAL_RC_OK
#define USE_LLDAL
#endif
#endif

#ifdef CBQDAEMON
#include "rsvp_cbq.h"
#else
#include "tc_cbq.h"
#endif
#include <netinet/altq_cbq.h>
#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <sys/ioctl.h>
#include <limits.h>

#ifdef USE_ATMPVC
#include <net/if_atm.h>		/* shadow2if() is defined in if_atm.h */
#define S2IF(n)	shadow2if(n)
#else
#define S2IF(n)	(n)
#endif

/*
 * Define local Macros
 */
#define CTL_PBANDWIDTH	2
#define NS_PER_MS	(1000000.0)
#define NS_PER_SEC	(NS_PER_MS*1000.0)
#define RM_FILTER_GAIN	5

#define	CBQ_DEVICE	"/dev/cbq"

#define ADMIT_FLOW_OK		0
#define ADMIT_FLOW_NOSVC	1
#define ADMIT_FLOW_NOBW		2
#define ADMIT_FLOW_SYSERR	3

#ifndef CBQDAEMON

typedef struct {
	int		in_if;		/* interface to which resv applies */
	cbq_flowinfo_t 	*cbq_handle;	/* cbq class handle */
	int		flags;
	u_short		nfilters;	/* Count filters for SE */
	SENDER_TSPEC	*s_tspec;	/* ptr to SENDER_TSPEC object */
	FLOWSPEC	*spec;		/* ptr to FLOWSPEC object */
	FILTER_SPEC	*filter;	/* ptr to FILTER_SPEC object */
	struct sockaddr_in	dest;
} tc_t;

#endif

/* 
 * External declarations (don't include math.h, log() conflicts)
 */
extern char           *fmt_spec();
extern double          pow(double, double);
extern double          fabs(double);

/* 
 * Forward references
 */
static int cbq_clear(char *ifname);
static cbq_ifinfo_t	*ifname2ifinfo(char *ifname);
static cbq_classinfo_t	*clname2clinfo(cbq_ifinfo_t *ifinfo, char *clname);
#ifndef CBQDAEMON
static int rsvp2cbq_filter(FilterSpec *filtp, Session *dest, cbq_filt_t *fl);
static void pr_filter(cbq_filt_t * filt);
static int admit_send_flow(tc_t *, FLOWSPEC *, SENDER_TSPEC *, int);
static void unadmit_send_flow(tc_t *);
static cbq_flowinfo_t *cbq_addflow(tc_t *);
static int cbq_delflow(cbq_flowinfo_t * flow_info);
static cbq_fltrinfo_t *cbq_addfilter(cbq_flowinfo_t * flinfo, cbq_filt_t * cfilter);
static int cbq_delfilter_all(cbq_flowinfo_t * flow_info);
static int cbq_delfilter(cbq_fltrinfo_t *);
extern Object_header	*copy_object(Object_header *);
#endif /* !CBQDAEMON */

/*
 * Static globals
 */

static cbq_ifinfo_t	*cbq_iflist = NULL;
static int		cbqfd = -1;

/*********************************************************************
 *  Following are CBQ-specific and are
 *  are called from above by RSVP.
 *
 *********************************************************************/

static cbq_ifinfo_t   *
ifname2ifinfo(char *ifname)
{
	cbq_ifinfo_t   *ifinfo;

	for (ifinfo = cbq_iflist; ifinfo; ifinfo = ifinfo->next)
		if (strcmp(ifinfo->ifname, ifname) == 0)
			return (ifinfo);
	return (NULL);
}

static cbq_classinfo_t	*
clname2clinfo(cbq_ifinfo_t *ifinfo, char *clname)
{
	cbq_classinfo_t	*clinfo;
	
	for (clinfo = ifinfo->cllist; clinfo; clinfo = clinfo->next)
		if (strcmp(clinfo->clname, clname) == 0)
			return (clinfo);
	return (NULL);
}

int
cbq_init()
{
	if (cbqfd >= 0)
	    close(cbqfd);

	cbq_cmds_init();

	if ((cbqfd = open(CBQ_DEVICE, O_RDWR)) < 0) {
		perror("CBQ open");
		return (-1);
	}

	/* read config file and initialize CBQ */
	return cbq_config();
}

int
cbq_add_if(char *ifname, long bandwidth, int is_sched, int is_wrr,
	   int efficient)
{
	cbq_ifinfo_t   *ifinfo;
	struct cbq_interface	cbqif;

	if ((ifinfo = calloc(1, sizeof(cbq_ifinfo_t))) == NULL)
		return (-1);
	ifinfo->ifname = strdup(ifname);
	ifinfo->cllist = NULL;
	ifinfo->bandwidth = bandwidth;
	ifinfo->nsPerByte =
		((1.0 / (double)ifinfo->bandwidth) * NS_PER_SEC * 8);
	ifinfo->use_cbq = is_sched && (cbqfd >= 0);
	ifinfo->is_wrr = is_wrr;
	ifinfo->is_efficient = efficient;
	ifinfo->enabled = 0;
	ifinfo->acc_enabled = 0;
	ifinfo->ifmtu = 0;

	/* Link the interface info structure */
	ifinfo->next = cbq_iflist;
	cbq_iflist = ifinfo;

	/* get mtu of the interface for default packet size */
	do {
		int s;
		struct ifreq ifr;

		if ((s = socket(AF_INET, SOCK_DGRAM, 0)) < 0)
			goto bad;
		strncpy(ifr.ifr_name, S2IF(ifname), sizeof ifr.ifr_name);
		if (ioctl(s, SIOCGIFMTU, (caddr_t)&ifr) < 0)
			goto bad;
		ifinfo->ifmtu = ifr.ifr_mtu;
	bad:
		if (ifinfo->ifmtu == 0)
			ifinfo->ifmtu = 512;	/* XXX */
		if (s >= 0)
			close(s);
	} while (0);

	/* attach cbq to the interface */
	memset((void *) &cbqif, 0, sizeof(cbqif));
	strncpy(cbqif.cbq_ifacename, S2IF(ifname), IFNAMSIZ);
	cbqif.cbq_ifacelen = strlen(S2IF(ifname));
	if (ifinfo->use_cbq &&
	    ioctl(cbqfd, CBQ_IF_ATTACH, &cbqif) < 0) {
		log(LOG_ERR, errno, "ioctl CBQ_IF_ATTACH");
		cbq_iflist = cbq_iflist->next;
		goto err_ret;
	}

	if (cbq_disable(ifname) || cbq_clear(ifname)) {
		cbq_iflist = cbq_iflist->next;
		goto err_ret;
	}
	return (0);

err_ret:
	if (ifinfo) {
		if (ifinfo->ifname)
			free(ifinfo->ifname);
		free(ifinfo);
	}
	return (-1);
}

int
cbq_enableall()
{
	cbq_ifinfo_t   *ifinfo;

	for (ifinfo = cbq_iflist; ifinfo; ifinfo = ifinfo->next)
		if (cbq_enable(ifinfo->ifname) < 0)
			return (-1);
	return (0);
}

int
cbq_enable(char *ifname)
{
	cbq_ifinfo_t   *ifinfo;
	struct cbq_interface cbqif;

	if ((ifinfo = ifname2ifinfo(ifname)) == NULL)
		return (-1);

	/*
	 * Enable the interface
	 */
	memset((void *) &cbqif, 0, sizeof(cbqif));
	strncpy(cbqif.cbq_ifacename, S2IF(ifname), IFNAMSIZ);
	cbqif.cbq_ifacelen = strlen(S2IF(ifname));
	if (ifinfo->use_cbq &&
	    ioctl(cbqfd, CBQ_ENABLE, &cbqif) < 0) {
		perror("ioctl CBQ_ENABLE");
		return (-1);
	}
	ifinfo->enabled = 1;

	log(LOG_INFO, 0, "CBQ enabled on interface %s (mtu:%d)",
	    ifname, ifinfo->ifmtu);

	return (0);
}

int
cbq_disableall()
{
	cbq_ifinfo_t   *ifinfo;
	int	error = 0;
	
	for (ifinfo = cbq_iflist; ifinfo; ifinfo = ifinfo->next)
		if (cbq_disable(ifinfo->ifname) < 0)
			error = -1;
	return (error);
}

int
cbq_disable(char *ifname)
{
	struct cbq_interface cbqif;
	cbq_ifinfo_t   *ifinfo;

	if ((ifinfo = ifname2ifinfo(ifname)) == NULL)
		return (-1);

	/* Disable the interface */
	memset((void *) &cbqif, 0, sizeof(cbqif));
	strncpy(cbqif.cbq_ifacename, S2IF(ifname), IFNAMSIZ);
	cbqif.cbq_ifacelen = strlen(S2IF(ifname));
	if (ifinfo->use_cbq &&
	    ioctl(cbqfd, CBQ_DISABLE, &cbqif) < 0) {
		perror("ioctl CBQ_DISABLE");
		return (-1);
	}
	ifinfo->enabled = 0;
	return (0);
}

int
cbq_acc_enable(char *ifname)
{
	struct cbq_interface cbqif;
	cbq_ifinfo_t   *ifinfo;

	if ((ifinfo = ifname2ifinfo(ifname)) == NULL)
		return (-1);
	/*
	 * Enable accounting
	 */
	memset((void *) &cbqif, 0, sizeof(cbqif));
	strncpy(cbqif.cbq_ifacename, S2IF(ifname), IFNAMSIZ);
	cbqif.cbq_ifacelen = strlen(S2IF(ifname));
	if (ifinfo->use_cbq &&
	    ioctl(cbqfd, CBQ_ACC_ENABLE, &cbqif) < 0) {
		perror("ioctl CBQ_ACC_ENABLE");
		return (-1);
	}
	ifinfo->acc_enabled = 1;
	return (0);
}

int
cbq_acc_disable(char *ifname)
{
	struct cbq_interface cbqif;
	cbq_ifinfo_t   *ifinfo;

	if ((ifinfo = ifname2ifinfo(ifname)) == NULL)
		return (-1);
	/*
	 * Disable accounting
	 */
	memset((void *) &cbqif, 0, sizeof(cbqif));
	strncpy(cbqif.cbq_ifacename, S2IF(ifname), IFNAMSIZ);
	cbqif.cbq_ifacelen = strlen(S2IF(ifname));
	if (ifinfo->use_cbq &&
	    ioctl(cbqfd, CBQ_ACC_DISABLE, &cbqif) < 0) 
		perror("ioctl CBQ_ACC_DISABLE");
	ifinfo->acc_enabled = 0;
	return (0);
}

int cbq_is_enabled(char *ifname)
{
	cbq_ifinfo_t   *ifinfo;

	if ((ifinfo = ifname2ifinfo(ifname)) == NULL)
		return (0);
	return (ifinfo->enabled);
}

int cbq_is_acc_enabled(char *ifname)
{
	cbq_ifinfo_t   *ifinfo;

	if ((ifinfo = ifname2ifinfo(ifname)) == NULL)
		return (0);
	return (ifinfo->acc_enabled);
}

static u_long
cbq_create_class(cbq_ifinfo_t * ifinfo, u_long parent_class,
	u_long borrow_class, u_int pri, int flags,
	u_int bandwidth, u_int maxdelay, u_int maxburst,
	u_int minburst, u_int av_pkt_size, u_int max_pkt_size)
{
	struct cbq_add_class cl_add;
	double          maxq, maxidle_s, maxidle, minidle,
			offtime, nsPerByte, ptime, cptime;
	double		z = (double)(1 << RM_FILTER_GAIN);
	double          g = (1.0 - 1.0 / z);
	double          f;
 	double		gton = pow(g, (double)maxburst);
 	double		gtom = pow(g, (double)minburst);
	double		maxrate;

 	/* Compute other class parameters */
	if (bandwidth == 0)
		f = 0.0001;	/* small enough? */
	else
		f = ((double) bandwidth / (double) ifinfo->bandwidth);

	if (av_pkt_size == 0) {	/* use default */
		av_pkt_size = ifinfo->ifmtu;
		if (av_pkt_size > MCLBYTES)	/* do what TCP does */
			av_pkt_size &= ~MCLBYTES;
	}
	else if (av_pkt_size > ifinfo->ifmtu)
		av_pkt_size = ifinfo->ifmtu;
	if (max_pkt_size == 0)	/* use default */
		max_pkt_size = ifinfo->ifmtu;
	else if (max_pkt_size > ifinfo->ifmtu)
		max_pkt_size = ifinfo->ifmtu;

        nsPerByte = ifinfo->nsPerByte / f;
	ptime = (double) av_pkt_size * (double)ifinfo->nsPerByte;
	maxrate = f * ((double)ifinfo->bandwidth / 8.0);
	cptime = ptime * (1.0 - f) / f;
#if 1 /* ALTQ */
	if (nsPerByte * (double)max_pkt_size > (double)INT_MAX) {
		/* this cause integer overflow in kernel! */
		if (bandwidth != 0)
			log(LOG_WARNING, 0, "warning: class is too slow!!");
		nsPerByte = (double)(INT_MAX / max_pkt_size);
	}
#endif
	if (IsDebug(DEBUG_CBQ)) {
		int packet_time;
		log(LOG_DEBUG, 0,
		    "cbq_create_class: maxburst=%d,minburst=%d,pkt_size=%d",
		    maxburst, minburst, av_pkt_size);
		log(LOG_DEBUG, 0,
		    "  nsPerByte=%.2f ns, link's nsPerByte=%.2f, f=%.3f",
		    nsPerByte, ifinfo->nsPerByte, f);
		packet_time = av_pkt_size * (int)nsPerByte / 1000;
		log(LOG_DEBUG, 0,
		    "  packet time=%d [us]", packet_time);
		if (maxburst * packet_time < 20000) {
			log(LOG_WARNING, 0,
			  "warning: maxburst smaller than timer granularity!");
			log(LOG_WARNING, 0,
			    "         maxburst=%d, packet_time=%d [us]",
			    maxburst, packet_time);
		}
	}
	maxidle = ((1.0 / f - 1.0) * ((1.0 - gton) / gton));
	maxidle_s = (1.0 - g);
	if (maxidle > maxidle_s)
		maxidle = ptime * maxidle;
	else
		maxidle = ptime * maxidle_s;
	if (IsDebug(DEBUG_CBQ))
		log(LOG_DEBUG, 0, "  maxidle=%.2f us", maxidle/1000.0);
	if (minburst)
		offtime = cptime * (1.0 + 1.0/(1.0 - g) * (1.0 - gtom) / gtom);
	else
		offtime = cptime;
	if (IsDebug(DEBUG_CBQ))
		log(LOG_DEBUG, 0, "  offtime=%.2f us", offtime/1000.0);
	maxidle = ((maxidle * 8.0) / nsPerByte) * pow(2, RM_FILTER_GAIN);
#if 1 /* ALTQ */
	/* scale offtime also */
 	offtime = (offtime * 8.0) / nsPerByte * pow(2, RM_FILTER_GAIN);
	minidle = -((double)max_pkt_size * (double)ifinfo->nsPerByte) / f;
#endif
	maxidle = maxidle / 1000.0;
	offtime = offtime / 1000.0;
	minidle = minidle / 1000.0;
	if (IsDebug(DEBUG_CBQ))
		log(LOG_DEBUG, 0, "  minidle=%.2f us", minidle/1000.0);
	/* adjust queue size when maxdelay is specified.
	   queue size should be relative to its share */
	if (maxdelay == 0)
		maxq = 30.0;
	else {
		maxq = ((double) maxdelay * NS_PER_MS) / (nsPerByte * av_pkt_size);
		if (maxq < 4) {
			log(LOG_WARNING, 0,
			    "warning: maxq (%d) is too small. set to %d",
			    (int)maxq, 4);
			maxq = 4;
		}
	}
	if (bandwidth == 0 && borrow_class == NULL_CLASS_HANDLE)
		/* filter out this class by setting queue size to zero */
		maxq = 0;
	if (IsDebug(DEBUG_CBQ)) {
		if ((int)maxq < maxburst)
			log(LOG_WARNING, 0,
			   "warning: maxq (%d) is smaller than maxburst(%d)",
			    (int)maxq, maxburst);
		else if (maxq > 100.0)
			log(LOG_WARNING, 0,
			    "warning: maxq %d too large", (int)maxq);
		log(LOG_DEBUG, 0, "  maxq=%d", (int)maxq);
	}

	/* create class */
	memset((void *) &cl_add, 0, sizeof(cl_add));
	strncpy(cl_add.cbq_iface.cbq_ifacename, S2IF(ifinfo->ifname),IFNAMSIZ);
	cl_add.cbq_iface.cbq_ifacelen = strlen(S2IF(ifinfo->ifname));
	cl_add.cbq_class.parent_class_handle = parent_class;
	cl_add.cbq_class.borrow_class_handle = borrow_class;
	cl_add.cbq_class.priority = pri;
	cl_add.cbq_class.nano_sec_per_byte = (u_int) nsPerByte;
	cl_add.cbq_class.maxq = (u_int) maxq;
	cl_add.cbq_class.maxidle = (u_int) fabs(maxidle);
	cl_add.cbq_class.minidle = (int)minidle;
	cl_add.cbq_class.offtime = (u_int) fabs(offtime);
	cl_add.cbq_class.pktsize = av_pkt_size;
	if (parent_class == NULL_CLASS_HANDLE) {
		if ((flags & CBQCLF_ROOTCLASS) == 0)
			flags |= CBQCLF_ROOTCLASS;
		if (ifinfo->is_wrr)
			flags |= CBQCLF_WRR;
		if (ifinfo->is_efficient)
			flags |= CBQCLF_EFFICIENT;
	}
	cl_add.cbq_class.flags = flags;

	if (ifinfo->use_cbq &&
	    ioctl(cbqfd, CBQ_ADD_CLASS, &cl_add) < 0) {
		perror("ioctl CBQ_ADD_CLASS");
		return (NULL_CLASS_HANDLE);
	}

	return cl_add.cbq_class_handle;
}

int
cbq_add_static_class(char *ifname, char *clname, char *parent_class,
     	char *borrow_class, u_int pri, int flags, u_int pbandwidth,
	u_int maxdelay, u_int maxburst, u_int minburst, u_int av_pkt_size,
	u_int max_pkt_size, cbq_tos_t admission_type,
	cbq_ol_t overlimit_action)
{
	cbq_ifinfo_t   *ifinfo;
	cbq_classinfo_t *parent_cl = NULL, *borrow_cl = NULL,
			*ctl_cl = NULL, *new_cl;
	u_long          cl_handle = NULL_CLASS_HANDLE;
	long            bandwidth;

	if ((ifinfo = ifname2ifinfo(ifname)) == NULL)
		return (-1);

	if (strcmp("ctl_class", clname) == 0)
		flags |= CBQCLF_CTLCLASS;
	else if (flags & CBQCLF_CTLCLASS)
		clname = "ctl_class";

	/* Get parent info */
	if (parent_class)
		if ((parent_cl = clname2clinfo(ifinfo, parent_class)) == NULL)
			return (-1);
	
	/* Get borrow info */
	if (borrow_class) {
		if ((borrow_cl = clname2clinfo(ifinfo, borrow_class)) == NULL)
			return (-1);
		if (borrow_cl != parent_cl) {
		    log(LOG_ERR, 0, "borrow class should be parent!");
		    return (-1);
		}
	}

	/* Create class */
	if (flags & CBQCLF_DEFCLASS) {
		/*
		 * if this is a default class and no ctl_class is defined,
		 * create a ctl_class.
		 */
		ctl_cl = clname2clinfo(ifinfo, "ctl_class");
		if (ctl_cl == NULL)
			/* reserve bandwidth for ctl_class */
			pbandwidth -= CTL_PBANDWIDTH;
	}

	bandwidth = ifinfo->bandwidth / 100 * pbandwidth;

	/* check bandwidth is lower than its parent */
	if (parent_cl != NULL && bandwidth > parent_cl->acinfo.bandwidth) {
		log(LOG_ERR, 0,
		    "bandwidth (%ld) greater then it parent's (%ld)",
		    bandwidth, parent_cl->acinfo.bandwidth);
		return (-1);
	}

	/* if average paket size isn't specified, set if mtu. */
	if (av_pkt_size == 0) {	/* use default */
		av_pkt_size = ifinfo->ifmtu;
		if (av_pkt_size > MCLBYTES)	/* do what TCP does */
			av_pkt_size &= ~MCLBYTES;
	}
	else if (av_pkt_size > ifinfo->ifmtu)
		av_pkt_size = ifinfo->ifmtu;

	if (max_pkt_size == 0)	/* use default */
		max_pkt_size = ifinfo->ifmtu;
	else if (max_pkt_size > ifinfo->ifmtu)
		max_pkt_size = ifinfo->ifmtu;

	/* instantiate class descriptor and init */
	if ((new_cl = calloc(1, sizeof(cbq_classinfo_t))) == NULL)
		return (-1);

	cl_handle = cbq_create_class(ifinfo,
		(parent_cl ? parent_cl->handle : NULL_CLASS_HANDLE),
		(borrow_cl ? borrow_cl->handle : NULL_CLASS_HANDLE),
	      	pri, flags, bandwidth, maxdelay,
		maxburst, minburst, av_pkt_size, max_pkt_size);
	if (cl_handle == NULL_CLASS_HANDLE)
		goto err_ret;

	new_cl->handle = cl_handle;
	new_cl->clname = strdup(clname);
	new_cl->handle = cl_handle;
	new_cl->parent_cl = parent_cl;
	new_cl->borrow_cl = borrow_cl;
	new_cl->ifinfo = ifinfo;

	/* Set pointers if a guranteed or predictive non-leaf class */
	switch (admission_type) {
	case CBQ_QOS_CNTR_LOAD:
	case CBQ_QOS_GUARANTEED:
#if 1
	case CBQ_QOS_PREDICTIVE:
	case CBQ_QOS_CNTR_DELAY:
#endif
		if (ifinfo->resv_class) {
			log(LOG_ERR, 0, "duplicate resv meta class");
			goto err_ret;
		}
		ifinfo->resv_class = new_cl;
		/* fall into... */
	case CBQ_QOS_NONE:
		new_cl->acinfo.type = admission_type;
		new_cl->acinfo.pri = pri;
		new_cl->acinfo.bandwidth = bandwidth;
		new_cl->acinfo.allocated = 0;
		new_cl->acinfo.maxdelay = maxdelay;
		new_cl->acinfo.maxburst = maxburst;
		new_cl->acinfo.minburst = minburst;
		new_cl->acinfo.av_pkt_size = av_pkt_size;
		new_cl->acinfo.max_pkt_size = max_pkt_size;
		break;
	default:
		goto err_ret;
	}

	/* link classinfo in list */
	new_cl->next = ifinfo->cllist;
	ifinfo->cllist = new_cl;

	/*
	 * If were adding the default traffic class and no ctl_class is
	 * defined, also add the ctl traffic class.  This is for RSVP,
	 * and IGMP packets.
	 */
	if ((flags & CBQCLF_DEFCLASS) && ctl_cl == NULL) {
		if (cbq_add_static_class(ifname, "ctl_class", parent_class,
			borrow_class, 6, CBQCLF_CTLCLASS, CTL_PBANDWIDTH,
			maxdelay, maxburst, minburst, av_pkt_size,
			max_pkt_size, admission_type, overlimit_action) < 0) {
			log(LOG_ERR, errno,
			    "can't create ctl_class");
		}
	}

	return (0);

err_ret:
	if (new_cl) {
		if (new_cl->clname)
			free(new_cl->clname);
		free(new_cl);
	}
	return (-1);
}

int
cbq_delete_static_class(char *ifname, char *clname)
{
	cbq_ifinfo_t   *ifinfo;
	cbq_classinfo_t *clinfo, *clprev;
	struct cbq_delete_class class_delete;

	if ((ifinfo = ifname2ifinfo(ifname)) == NULL)
		return (-1);

	/* remove class info from the interface */
	if ((clinfo = clprev = ifinfo->cllist) == NULL)
		return (-1);
	if (strcmp(clinfo->clname, clname) == 0) {
		ifinfo->cllist = clinfo->next;
	}
	else while ((clinfo = clprev->next) != NULL) {
		if (strcmp(clinfo->clname, clname) == 0) {
			clprev->next = clinfo->next;
			break;
		}
		clprev = clinfo;
	}

#ifndef CBQDAEMON
	/* delete flowinfo associated with this class */
	while (clinfo->acinfo.flow_list != NULL)
		if (cbq_delflow(clinfo->acinfo.flow_list) < 0) {
			log(LOG_ERR, errno, "can't delete flow");
			break;
		}
#endif

	/* delete class from kernel */
	if (clinfo->handle != NULL_CLASS_HANDLE) {
		memset((void *) &class_delete, 0, sizeof(class_delete));
		strncpy(class_delete.cbq_iface.cbq_ifacename,
			S2IF(ifname), IFNAMSIZ);
		class_delete.cbq_iface.cbq_ifacelen = strlen(S2IF(ifname));
		class_delete.cbq_class_handle = clinfo->handle;
		if (ifinfo->use_cbq &&
		    ioctl(cbqfd, CBQ_DEL_CLASS, &class_delete) < 0) {
			perror("ioctl CBQ_DEL_CLASS");
			return (-1);
		}
	}

	if (clinfo->clname)
		free(clinfo->clname);
	free(clinfo);
	return (0);
}

int
cbq_add_static_filter(char *ifname, char *clname, cbq_filt_t *filt)
{
	cbq_ifinfo_t   *ifinfo;
	cbq_classinfo_t *clinfo = NULL;
	struct cbq_add_filter fltr_add;

	if ((ifinfo = ifname2ifinfo(ifname)) == NULL)
		return (-1);

	if ((clinfo = clname2clinfo(ifinfo, clname)) == NULL)
		return (-1);

	memset((void *) &fltr_add, 0, sizeof(fltr_add));
	strncpy(fltr_add.cbq_iface.cbq_ifacename, S2IF(ifname), IFNAMSIZ);
	fltr_add.cbq_iface.cbq_ifacelen = strlen(S2IF(ifname));
	fltr_add.cbq_class_handle = clinfo->handle;
	fltr_add.cbq_filter = *filt;

	if (IsDebug(DEBUG_CBQ)) {
	        struct in_addr in_addr;

		fprintf(stderr, "%s: CBQ_ADD_FILTER, class.%lx\n",
			ifname, fltr_add.cbq_class_handle);
		in_addr.s_addr = fltr_add.cbq_filter.daddr;
		fprintf(stderr, "Filter Destination Address: %s mask 0x%x\n",
			inet_ntoa(in_addr), ntoh32(fltr_add.cbq_filter.dmask));
		fprintf(stderr, "Filter Destination Port: %d\n",
			ntoh16(fltr_add.cbq_filter.dport));
		in_addr.s_addr = fltr_add.cbq_filter.saddr;
		fprintf(stderr, "Filter Source Address: %s mask 0x%x\n",
			inet_ntoa(in_addr), ntoh32(fltr_add.cbq_filter.smask));
		fprintf(stderr, "Filter Source Port: %d\n",
			ntoh16(fltr_add.cbq_filter.sport));
		fprintf(stderr, "Filter IP Protocol: %d\n",
			fltr_add.cbq_filter.proto);
		fprintf(stderr, "Filter TOS 0x%x (mask 0x%x)\n",
			fltr_add.cbq_filter.tos, fltr_add.cbq_filter.tosmask);
		fprintf(stderr, "\n");
	}

	if (ifinfo->use_cbq &&
	    ioctl(cbqfd, CBQ_ADD_FILTER, &fltr_add) < 0) {
		perror("ioctl CBQ_ADD_FILTER");
		return (-1);
	}
	return (0);
}

static int
cbq_clear(char *ifname)
{
	struct cbq_interface 	cbqif;
	cbq_ifinfo_t  		*ifinfo;

	if ((ifinfo = ifname2ifinfo(ifname)) == NULL)
		return (-1);
		
	/* clear the interface */
	memset((void *) &cbqif, 0, sizeof(cbqif));
	strncpy(cbqif.cbq_ifacename, S2IF(ifname), IFNAMSIZ);
	cbqif.cbq_ifacelen = strlen(S2IF(ifname));
	if (ifinfo->use_cbq &&
	    ioctl(cbqfd, CBQ_CLEAR_HIERARCHY, &cbqif) < 0) {
		perror("ioctl CBQ_CLEAR_HIERARCHY");
		return (-1);
	}
	return (0);
}

int
cbq_destroyall(void)
{
	cbq_ifinfo_t		*ifinfo, *ifnext;
	cbq_classinfo_t		*clinfo, *clnext;
	struct cbq_interface	cbqif;

	if (cbqfd < 0) {
		log(LOG_ERR, 0, "cbq_destroyall: CBQ device is not opened");
		return (-1);
	}

	for (ifinfo = cbq_iflist; ifinfo; ifinfo = ifnext) {

		(void) cbq_disable(ifinfo->ifname);
		
		/* free all classes */
		for (clinfo = ifinfo->cllist; clinfo; clinfo = clnext) {
#ifndef CBQDAEMON
			/* delete flowinfo associated with this class */
			while (clinfo->acinfo.flow_list != NULL)
				if (cbq_delflow(clinfo->acinfo.flow_list) < 0) {
					log(LOG_ERR, errno,
					    "can't delete flow");
					break;
				}
#endif
			clnext = clinfo->next;
			if (clinfo->clname)
				free(clinfo->clname);
			free(clinfo);
		}

		(void)cbq_clear(ifinfo->ifname);

		memset((void *) &cbqif, 0, sizeof(cbqif));
		strncpy(cbqif.cbq_ifacename, S2IF(ifinfo->ifname), IFNAMSIZ);
		cbqif.cbq_ifacelen = strlen(S2IF(ifinfo->ifname));
			
		if (ifinfo->use_cbq &&
		    ioctl(cbqfd, CBQ_IF_DETACH, &cbqif) < 0)
			perror("ioctl CBQ_IF_DETACH");

		ifnext = ifinfo->next;
		if (ifinfo->ifname)
			free(ifinfo->ifname);
		free(ifinfo);
	}

	cbq_iflist = NULL;
	if (cbqfd >= 0 && close(cbqfd) < 0)
		perror("close cbq");
	cbqfd = -1;
	return (0);
}


#ifndef CBQDAEMON  /* only for RSVP */
/************************************************************************
 *
 *	Interface routines to call the CBQ kernel functions
 * 
 ************************************************************************/

/* Note: these calls use the address of the tc_t block as a parameter
 *	to carry the interface number.  This block actually contains the
 *	the other parameters too, but we leave the parameters explicit to
 *	make the code exactly parallel the Functional Specification.
 */
void
TC_init(int OIf)
{
	int i;
	static once = FALSE;

	if (once)
		return;
	once = TRUE;
	if(cbq_init() < 0) {
		log(LOG_ERR, 0, "Can't initialize CBQ!");
	}

        for (i = 0; i < if_num; i++) {
		cbq_ifinfo_t   *ifinfo;

		if (IsNumAPI(i))
			continue;
		if_vec[i].if_up = 0;

		for (ifinfo = cbq_iflist; ifinfo; ifinfo = ifinfo->next)
			if (strcmp(ifinfo->ifname, if_vec[i].if_name) == 0) {
				if (ifinfo->resv_class != NULL) {
					if_vec[i].if_up = 1;
				}
				else
					log(LOG_WARNING, 0,
					    "%s: No Class defined for RSVP",
					    ifinfo->ifname);
			}
	}
}

/*
 * TC_AddFlowspec(): Call the kernel to make reservation for a flow.
 * 	It checks admission control, and returns a handle for the
 *	reservation, or -1 if an error.
 */
u_long
TC_AddFlowspec(int OIf, FLOWSPEC *spec, SENDER_TSPEC *stspec,
#ifdef USE_LLDAL
	       				ADSPEC *adspp,
#endif
					int flags, FLOWSPEC **fwd_specpp)
{
	cbq_flowinfo_t 	*cbq_handle = NULL;
	tc_t		*tc_handle = NULL;

	rsvp_errno = 0;
	*fwd_specpp = NULL;

	if(IsNumAPI(OIf))
		return((u_long)NULL);

	/* Handle outgoing flow */
	rsvp_errno = 0;
	if (!if_vec[OIf].if_up) {
		rsvp_errno = Set_Errno(RSVP_Err_TC_SYS_ERROR, 0);
		return(TC_ERROR);
	}
		
	/* Create reservation control block and init */
	if((tc_handle = (tc_t*)malloc(sizeof(tc_t))) == NULL) {
		rsvp_errno = Set_Errno(RSVP_Err_TC_SYS_ERROR, 0);
		return(TC_ERROR);
	}
	tc_handle->in_if = OIf;
	tc_handle->flags = flags;
	tc_handle->cbq_handle = NULL;
	tc_handle->spec = copy_spec(spec);
	tc_handle->s_tspec = stspec ? copy_tspec(stspec) : NULL;
		
	/* Can we admit send flow? */
	switch(admit_send_flow(tc_handle, spec, stspec, flags)){
	case ADMIT_FLOW_OK:
		if((cbq_handle = cbq_addflow(tc_handle)) == NULL) {
			(void)unadmit_send_flow(tc_handle);
			rsvp_errno = Set_Errno(RSVP_Err_TC_SYS_ERROR, 0);
			goto err_ret;
		}
		tc_handle->cbq_handle = cbq_handle;
		return((u_long)tc_handle);
			
	case ADMIT_FLOW_NOSVC:
		rsvp_errno = Set_Errno(RSVP_Err_TC_ERROR,
				       RSVP_Erv_No_Serv);
		goto err_ret;
		
	case ADMIT_FLOW_NOBW:
		rsvp_errno = Set_Errno(RSVP_Err_ADMISSION, 
				       RSVP_Erv_Bandwidth);
		goto err_ret;

	case ADMIT_FLOW_SYSERR:
	default:
		rsvp_errno = Set_Errno(RSVP_Err_TC_SYS_ERROR, 0);
		goto err_ret;
	}
	
 err_ret:
    	if(tc_handle) {
    		if(tc_handle->spec) free(tc_handle->spec);
    		if(tc_handle->s_tspec) free(tc_handle->s_tspec);
    		free(tc_handle);
    	}
    	return(TC_ERROR);
}

/*
 * TC_DelFlowspec(): This routine deletes flow for specified handle
 */

int
TC_DelFlowspec(int OIf, u_long rhandle) 
{
	tc_t		*tch;

	if (rhandle == TC_ERROR)
		return TC_ERROR;
	if((tch = (tc_t*)rhandle) == NULL || IsNumAPI(OIf))
		return(TC_OK);
		
	/* Handle outgoing flow */
	unadmit_send_flow(tch);
	if(cbq_delflow(tch->cbq_handle) < 0) {
		if(!rsvp_errno) 
			rsvp_errno = Set_Errno(RSVP_Err_TC_SYS_ERROR, 0);
		return TC_ERROR;
	}
	if(tch->spec) free(tch->spec);
	if(tch->s_tspec) free(tch->s_tspec);
	free(tch);
	return TC_OK;
}

/*
 * TC_ModFlowspec(): Modifies a flowspec of a given flow.
 */
int
TC_ModFlowspec(int OIf, u_long rhandle,
	 		FLOWSPEC *specp, SENDER_TSPEC *stspecp,
#ifdef USE_LLDAL
	       				ADSPEC *adspecp,
#endif
					int flags, FLOWSPEC **fwd_specpp)
{
	tc_t		*tc_handle;
	tc_t		tc_ohandle;
	cbq_flowinfo_t	*cbq_handle;
	cbq_fltrinfo_t	*fltr, *sfltr, *saved_filters;
	cbq_ifinfo_t	*ifinfo;

	/*
	 * modification by kjc
	 *   - save and restore the filters.
	 *   - leave the original reservation even if TC_ModFlowspec fails.
	 */

	/* NB: This is only safe if rsvpd remains single threaded as it
	 * is now.  If it becomes MT, then need a lock on admission control */

	rsvp_errno = 0;
	*fwd_specpp = NULL;

	if((tc_handle = (tc_t*)rhandle) == NULL || IsNumAPI(OIf))
		return (TC_OK);

	/* Handle outgoing flow */
	cbq_handle = tc_handle->cbq_handle;
	ifinfo = cbq_handle->clinfo->ifinfo;

	/*
	 * first, check if we can admit this update.
	 * save old flowspec in case of admission failure.
	 */
	tc_ohandle = *tc_handle;

	unadmit_send_flow(tc_handle);

	tc_handle->flags = flags;
	tc_handle->spec = copy_spec(specp);
	tc_handle->s_tspec = stspecp ? copy_tspec(stspecp) : NULL;
	
	/* Can we admit new send flow? */
	switch(admit_send_flow(tc_handle, specp, stspecp, flags)){
	case ADMIT_FLOW_OK:
		break;
	case ADMIT_FLOW_NOSVC:
		rsvp_errno = Set_Errno(RSVP_Err_TC_ERROR, RSVP_Erv_No_Serv);
		break;
	case ADMIT_FLOW_NOBW:
		rsvp_errno = Set_Errno(RSVP_Err_ADMISSION, RSVP_Erv_Bandwidth);
		break;
	case ADMIT_FLOW_SYSERR:
	default:
		rsvp_errno = Set_Errno(RSVP_Err_TC_SYS_ERROR, 0);
		break;
	}

	if (rsvp_errno == 0) {
		/* update seems ok, save filters */
		saved_filters = NULL;
		for (fltr = cbq_handle->fltr_list; fltr; fltr = fltr->next) {
			if ((sfltr = calloc(1, sizeof(cbq_fltrinfo_t))) == NULL) {
				log(LOG_ERR, errno, "can't alloc filters");
				rsvp_errno = Set_Errno(RSVP_Err_TC_SYS_ERROR, 0);
				while ((sfltr = saved_filters) != NULL) {
					saved_filters = sfltr->next;
					free(sfltr);
				}
				break;
			}
			sfltr->filter = fltr->filter;	/* save filter value */
			sfltr->next = saved_filters;
			saved_filters = sfltr;
		}
	}
	if (rsvp_errno) {
		/* can't update.  restore the original flowspec */
		if(tc_handle->spec) free(tc_handle->spec);
		if(tc_handle->s_tspec) free(tc_handle->s_tspec);

		tc_handle->flags = tc_ohandle.flags;
		tc_handle->spec = tc_ohandle.spec;
		tc_handle->s_tspec = tc_ohandle.s_tspec;
		if (admit_send_flow(tc_handle, tc_handle->spec,
				    tc_handle->s_tspec, tc_handle->flags)
		    != ADMIT_FLOW_OK) {
			/* ick! can't restore the original flow! */
			log(LOG_ERR, 0, "can't restore the original flow!");
		}
		return (TC_ERROR);
	}

	/*
	 * now we have decided to commit the update.
	 * errors should not happen from now on, since we are no longer
	 * able to restore the original flowspec.
	 */
	if(tc_ohandle.spec) free(tc_ohandle.spec);
	if(tc_ohandle.s_tspec) free(tc_ohandle.s_tspec);

	if(cbq_delflow(cbq_handle) < 0) {
		log(LOG_ERR, errno, "can't delete flowspec!");
		(void)unadmit_send_flow(tc_handle);
		if(!rsvp_errno) 
			rsvp_errno = Set_Errno(RSVP_Err_TC_SYS_ERROR, 0);
		return (TC_ERROR);
	}

	if((cbq_handle = cbq_addflow(tc_handle)) == NULL) {
		log(LOG_ERR, errno, "can't add flowspec!");
		(void)unadmit_send_flow(tc_handle);
		tc_handle->cbq_handle = NULL;
		rsvp_errno = Set_Errno(RSVP_Err_TC_SYS_ERROR, 0);
		return (TC_ERROR);
	}
	tc_handle->cbq_handle = cbq_handle;

	/* restore filters */
	while ((sfltr = saved_filters) != NULL) {
		cbq_addfilter(cbq_handle, &sfltr->filter);
		saved_filters = sfltr->next;
		free(sfltr);
	};

	return (TC_OK);
}

/*
 * TC_AddFilter(): Adds a filter for an existing flow.
 *
 *	Returns fhandle or TC_ERROR.
 */
u_long
TC_AddFilter(int OIf, u_long rhandle, Session *dest, FILTER_SPEC *filtp)
{
	cbq_filt_t      cfilter;
	tc_t		*tch = (tc_t*)rhandle;
	cbq_fltrinfo_t	*fh;

	if(tch == NULL || IsNumAPI(tch->in_if) || !if_vec[tch->in_if].if_up) {
		/* CBQ does not create kernel state for incoming flows */
		return TC_OK;
	}

	rsvp_errno = 0;
		
	/* Add or modify the filter */
	if(rsvp2cbq_filter(filtp, dest, &cfilter) < 0 ||
	   (fh = cbq_addfilter(tch->cbq_handle, &cfilter)) == NULL) {
		if(!rsvp_errno) 
			rsvp_errno = RSVP_Err_TC_SYS_ERROR;
		return TC_ERROR;
	}
	return (u_long)fh;
}


/*
 * TC_DelFilter(): Deletes existing filter.
 */
int
TC_DelFilter(int OIf, u_long fhandle)
{
	cbq_fltrinfo_t *fh = (cbq_fltrinfo_t *)fhandle;

	if(fh == NULL || IsNumAPI(OIf) || !if_vec[OIf].if_up)
		return TC_OK;

	rsvp_errno = 0;
	if(cbq_delfilter(fh) < 0) {
		if(!rsvp_errno) 
			rsvp_errno = RSVP_Err_TC_SYS_ERROR;
		return TC_ERROR;
	}
	return TC_OK;
}


#define In_Obj(x, y) ((Object_header *)(x) <= Next_Object((Object_header *)(y)))

/*
 * TC_Advertise(): Given existing OPWA ADSPEC, return a new updated object.
 */
ADSPEC *
TC_Advertise(int OIf, ADSPEC * old_asp, int flags)
{
	ADSPEC		*new_asp = copy_adspec(old_asp);
	/*	We know we will not expand the ADSPEC, so start by
	 *	just making a straight copy.
	 */
	IS_main_hdr_t	*mhp = (IS_main_hdr_t *) Obj_data(new_asp);
	genparm_parms_t *gpp = (genparm_parms_t *)(mhp +1);
	IS_serv_hdr_t	*shp, *lastshp;
	float32_t	hop_bw;
	u_int32_t	hop_latency, hop_mtu;
	cbq_ifinfo_t	*ifinfo;

	if (if_vec[OIf].if_up == 0 || (flags & ADVERTF_NonRSVP)) {
		Set_Break_Bit(&gpp->gen_parm_hdr);
	}
	else {
		/* get interface info, to over-ride interface defaults */
		if ((ifinfo = ifname2ifinfo(if_vec[OIf].if_name)) != NULL)
			if_vec[OIf].if_path_bw =
				ifinfo->resv_class->acinfo.bandwidth
				- ifinfo->resv_class->acinfo.allocated;
	}
	
	gpp->gen_parm_hopcnt++;

	hop_bw = (if_vec[OIf].if_path_bw)?if_vec[OIf].if_path_bw:
							(TC_DFLT_PATH_BW);
	gpp->gen_parm_path_bw = MIN(gpp->gen_parm_path_bw, hop_bw);

	hop_latency = (if_vec[OIf].if_min_latency)?if_vec[OIf].if_min_latency:
							TC_DFLT_MIN_LATENCY;
	gpp->gen_parm_min_latency = MIN(gpp->gen_parm_min_latency, hop_latency);
	
	hop_mtu = (if_vec[OIf].if_path_mtu)?if_vec[OIf].if_path_mtu: 
							TC_DFLT_MTU;
	gpp->gen_parm_composed_MTU = MIN(gpp->gen_parm_composed_MTU, hop_mtu);
	
	if (if_vec[OIf].if_up == 0 || (flags & ADVERTF_NonRSVP)) {
		/* We KNOW that we are not capable of traffic control, so we
		 * set Break bit in each service.  We DO NOT set the Break
		 * bit in the generic parameters, because we DO understand
		 * ADSPECs, unless previous hop was not RSVP-capable.
		 */
		shp = Next_Serv_Hdr((IS_serv_hdr_t *)gpp);
		lastshp  = (IS_serv_hdr_t *) Next_Main_Hdr(mhp);
		if (!In_Obj(lastshp, new_asp)) {
			/* Internal length falls outside object...
			 *  Increment error counter
			 */
			free(new_asp);
			return(NULL);
		}
		while (shp < lastshp) {
			switch(shp->issh_service) {
			case CONTROLLED_LOAD_SERV:
				shp->issh_flags^=ISSH_BREAK_BIT;
				break;
			default:
				Set_Break_Bit(shp);
				break;
			}
			shp = Next_Serv_Hdr(shp);
		}
	}
	return(new_asp);
}


static cbq_flowinfo_t *
cbq_addflow(tc_t *tch)
{
	IS_specbody_t		*z_specp;
	cbq_ifinfo_t  		*ifinfo;
	cbq_acinfo_t   		*acinfo;
	cbq_classinfo_t 	*clinfo;
	u_long          	cl_handle;
	cbq_flowinfo_t 		*new_fl;
	u_long			bandwidth;

	assert(tch);
	assert(Obj_CType(tch->spec) == ctype_FLOWSPEC_Intserv0);
	z_specp = &(tch->spec->flow_body);
	bandwidth = z_specp->spec_u.CL_spec.CLspec_r * 8;

	/* get interface info */
	if ((ifinfo = ifname2ifinfo(if_vec[tch->in_if].if_name)) == NULL)
		return NULL;
	if ((clinfo = ifinfo->resv_class) == NULL)
		return NULL;
	acinfo = &(ifinfo->resv_class->acinfo);
	
	if ((new_fl = calloc(1, sizeof(cbq_flowinfo_t))) == NULL)
		return NULL;
	
	/* create a leaf class */
	cl_handle = cbq_create_class(ifinfo, clinfo->handle, clinfo->handle,
	        acinfo->pri, 0, bandwidth, acinfo->maxdelay,
		acinfo->maxburst, acinfo->minburst, acinfo->av_pkt_size,
	        acinfo->max_pkt_size);
	if (cl_handle == NULL_CLASS_HANDLE) {
		free(new_fl);
		return NULL;
	}

	new_fl->clinfo = clinfo;
/*XXX	new_fl->type = z_specp->spec_type;*/
	new_fl->bandwidth = bandwidth;
	new_fl->handle = cl_handle;
	new_fl->next = acinfo->flow_list;
	acinfo->flow_list = new_fl;
	return new_fl;
}

static int
cbq_delflow(cbq_flowinfo_t * flow_info)
{
	cbq_classinfo_t *clinfo = flow_info->clinfo;
	cbq_ifinfo_t	*ifinfo = clinfo->ifinfo;
	struct cbq_delete_class	class_delete;
	struct cbq_flowinfo	*flow_list;

	/* delete any outstanding filters on class for this flow */
	if (cbq_delfilter_all(flow_info) < 0)
		return (-1);

	if (flow_info->handle != NULL_CLASS_HANDLE) {
		/* delete class from kernel */
		memset((void *) &class_delete, 0, sizeof(class_delete));
		strncpy(class_delete.cbq_iface.cbq_ifacename,
			S2IF(ifinfo->ifname), IFNAMSIZ);
		class_delete.cbq_iface.cbq_ifacelen = strlen(S2IF(ifinfo->ifname));
		class_delete.cbq_class_handle = flow_info->handle;
		if (ifinfo->use_cbq &&
		    ioctl(cbqfd, CBQ_DEL_CLASS, &class_delete) < 0) {
			perror("ioctl CBQ_DEL_CLASS");
			return (-1);
		}
	}

	/* remove from list of flows for this reservation class */
	if (clinfo->acinfo.flow_list == flow_info) {
		clinfo->acinfo.flow_list = flow_info->next;
	} else {
		flow_list = clinfo->acinfo.flow_list;
		while (flow_list->next) {
			if (flow_list->next == flow_info) {
				flow_list->next = flow_info->next;
				break;
			}
			flow_list = flow_list->next;
		}
	}
	free(flow_info);
	return 0;
}

static cbq_fltrinfo_t*
cbq_addfilter(cbq_flowinfo_t *flinfo, cbq_filt_t *cfilter)
{
	cbq_ifinfo_t   *ifinfo = flinfo->clinfo->ifinfo;
	struct cbq_add_filter fltr_add;
	cbq_fltrinfo_t *fltr_info;

	if (IsDebug(DEBUG_CBQ))
		pr_filter(cfilter);

	if ((fltr_info = calloc(1, sizeof(cbq_fltrinfo_t))) == NULL)
		return NULL;

	memset((void *) &fltr_add, 0, sizeof(fltr_add));
	strncpy(fltr_add.cbq_iface.cbq_ifacename, S2IF(ifinfo->ifname), IFNAMSIZ);
	fltr_add.cbq_iface.cbq_ifacelen = strlen(S2IF(ifinfo->ifname));
	fltr_add.cbq_class_handle = flinfo->handle;
	fltr_add.cbq_filter = *cfilter;

	if (IsDebug(DEBUG_CBQ)) {
	        struct in_addr in_addr;
		fprintf(stderr, "%s: CBQ_ADD_FILTER, class.%lx\n",
			ifinfo->ifname, fltr_add.cbq_class_handle);
		in_addr.s_addr = fltr_add.cbq_filter.daddr;
		fprintf(stderr, "Filter Destination Address: %s\n",
			inet_ntoa(in_addr));
		fprintf(stderr, "Filter Destination Port: %d\n",
			ntoh16(fltr_add.cbq_filter.dport));
		in_addr.s_addr = fltr_add.cbq_filter.saddr;
		fprintf(stderr, "Filter Source Address: %s\n",
			inet_ntoa(in_addr));
		fprintf(stderr, "Filter Source Port: %d\n",
			ntoh16(fltr_add.cbq_filter.sport));
		fprintf(stderr, "Filter IP Protocol: 0x%x\n",
			fltr_add.cbq_filter.proto);
		fprintf(stderr, "\n");
	}

	if (ifinfo->use_cbq &&
	    ioctl(cbqfd, CBQ_ADD_FILTER, &fltr_add) < 0) {
		perror("ioctl CBQ_ADD_FILTER");
		free(fltr_info);
		return NULL;
	}

	fltr_info->handle = fltr_add.cbq_filter_handle;
	fltr_info->filter = *cfilter;
	fltr_info->flow = flinfo;

	fltr_info->next = flinfo->fltr_list;
	flinfo->fltr_list = fltr_info;

	return (fltr_info);
}

static int
cbq_delfilter_all(cbq_flowinfo_t * flow_info)
{
	while (flow_info->fltr_list != NULL)
		if (cbq_delfilter(flow_info->fltr_list) < 0)
			return (-1);
	return (0);
}

static int 
cbq_delfilter(cbq_fltrinfo_t *fltr)
{
	struct cbq_delete_filter fltr_delete;
	cbq_flowinfo_t *flinfo = fltr->flow;
	cbq_ifinfo_t   *ifinfo = flinfo->clinfo->ifinfo;
	cbq_fltrinfo_t *fltr_prev;
	
	memset((void *) &fltr_delete, 0, sizeof(fltr_delete));
	strncpy(fltr_delete.cbq_iface.cbq_ifacename,
		S2IF(ifinfo->ifname), IFNAMSIZ);
	fltr_delete.cbq_iface.cbq_ifacelen = strlen(S2IF(ifinfo->ifname));
	fltr_delete.cbq_filter_handle = fltr->handle;
	if (ifinfo->use_cbq &&
	    ioctl(cbqfd, CBQ_DEL_FILTER, &fltr_delete) < 0) {
		perror("ioctl CBQ_DEL_FILTER");
		return (-1);
	}

	/* remove from the filter list */
	if (flinfo->fltr_list == fltr)
		flinfo->fltr_list = fltr->next;
	else {
		fltr_prev = flinfo->fltr_list;
		while (fltr_prev->next != NULL) {
			if (fltr_prev->next == fltr) {
				fltr_prev->next = fltr->next;
				break;
			}
			fltr_prev = fltr_prev->next;
		}
	}
	free(fltr);
	return (0);
}

#ifndef filt4_srcaddr
/* for isi rsvp rel4.1 compatibility */
#define filt4			filt_u.filt_ipv4
#define filt4_srcaddr		filt4.filt_ipaddr
#define filt4_srcport		filt4.filt_port
#endif

static int
rsvp2cbq_filter(FilterSpec *filtp, Session *dest, cbq_filt_t *flp)
{
	cbq_filt_t 		fl;

	memset((void *)&fl, 0, sizeof(cbq_filt_t));

	switch (Obj_CType(dest->d_session)) {
	case ctype_SESSION_ipv4:
		fl.family = AF_INET;
		fl.daddr  = dest->d_session->sess4_addr.s_addr;
		fl.dmask  = 0xffffffff;
		fl.dport  = dest->d_session->sess4_port;
		fl.proto  = dest->d_session->sess4_prot;

		if (filtp == NULL) {
			/* IPv4 wildcard filter */
			fl.saddr = 0;
			fl.sport = 0;
			
		} else switch (Obj_CType(filtp)) {
		case ctype_FILTER_SPEC_ipv4:	/* IPv4 FILTER_SPEC */
			fl.saddr = filtp->filt4_srcaddr.s_addr;
			fl.smask = 0xffffffff;
			fl.sport = filtp->filt4_srcport;
			break;

		case ctype_FILTER_SPEC_ipv4GPI:	/* IPv4/GPI FILTER_SPEC */
			fl.dport  = 0; /* we don't need dst port */
			fl.saddr  = filtp->filtgpi4_srcaddr.s_addr;
			fl.smask  = 0xffffffff;
			fl.gpi    = filtp->filtgpi4_srcgpi;
			break;
			
		default:
			rsvp_errno = Set_Errno(RSVP_Err_UNKNOWN_CTYPE,
					       Obj_CType(filtp));
			return (-1);
		}
		
		break;

#ifdef USE_IPV6			
	case ctype_SESSION_ipv6:
		fl.family   = AF_INET6;
		fl.daddr6   = dest->d_session->sess6_addr;
		memset(&fl.dmask6, 0xff, sizeof(struct in6_addr));
		fl.dport6   = dest->d_session->sess6_port;
		fl.proto6   = dest->d_session->sess6_prot;

		if (filtp == NULL) {
			/* IPv6 wildcard */
			memset(&fl.saddr6, 0, sizeof(struct in6_addr));
			fl.sport6   = 0;
			
		} else switch (Obj_CType(filtp)) {
		case ctype_FILTER_SPEC_ipv6:	/* IP6 FILTER_SPEC */
			memcpy(&fl.saddr6, filtp->filt6_srcaddr,
			       sizeof(struct in6_addr));
			memset(&fl.smask6, 0xff, sizeof(struct in6_addr));
			fl.sport6   = filtp->filt6_srcport;
			break;
			
#ifdef notyet /* not yet supported in rsvp rel4.2a3 */
		case ctype_FILTER_SPEC_ipv6FL: /* IP6 Flow-label FILTER_SPEC */
			fl.dport6 = 0;  /* we don't need dst port */
			memcpy(&fl.saddr6, filtp->filtfl6_srcaddr,
			       sizeof(struct in6_addr));
			memset(&fl.smask6, 0xff, sizeof(struct in6_addr));
			fl.flowlabel6 = filtp->filtfl6_srcflowlable
				& htonl(0x000fffff);
			break;
#endif
		case ctype_FILTER_SPEC_ipv6GPI:	/* IPv6/GPI FILTER_SPEC */
			fl.dport6 = 0;  /* we don't need dst port */
			memcpy(&fl.saddr6, filtp->filtgpi6_srcaddr,
			       sizeof(struct in6_addr));
			memset(&fl.smask6, 0xff, sizeof(struct in6_addr));
			fl.gpi6     = filtp->filtgpi6_srcgpi;
			break;

		default:
			rsvp_errno = Set_Errno(RSVP_Err_UNKNOWN_CTYPE,
					       Obj_CType(filtp));
			return (-1);
		}

		break;
#endif /* USE_IPV6 */

#ifdef notyet
	/* not supported in rsvp rel4.2a3 */
	case ctype_SESSION_ipv4GPI:
	case ctype_SESSION_ipv6GPI:
#endif
	default:
		rsvp_errno = Set_Errno(RSVP_Err_UNKNOWN_CTYPE,
				       Obj_CType(dest->d_session));
		return (-1);
		break;
	}
	
	*flp = fl;
	return (0);
}

static void
pr_filter(cbq_filt_t * filt)
{
        struct in_addr in_addr;

	in_addr.s_addr = filt->daddr;
	log(LOG_INFO, 0, "Dest Addr = %s, Dest Port = %x",
	    inet_ntoa(in_addr), filt->dport);
	in_addr.s_addr = filt->saddr;
	log(LOG_INFO, 0, "Source Addr = %x, Source Port = %x",
	    inet_ntoa(in_addr), filt->sport);
	log(LOG_INFO, 0, "Protocol = %x", filt->proto);
}

static int
admit_send_flow(tc_t *tch, FLOWSPEC *spec, SENDER_TSPEC *stspec, int flags)
{
	cbq_ifinfo_t		*ifinfo;
	cbq_acinfo_t		*acinfo;
	cbq_classinfo_t 	*clinfo;
	IS_specbody_t		*z_specp;
	IS_serv_hdr_t		*sp;
	long			req_bandwidth;

	assert(tch);
	
	if ((ifinfo = ifname2ifinfo(if_vec[tch->in_if].if_name)) == NULL)
		return ADMIT_FLOW_SYSERR;

	/* Get relevant information from flow spec */
	if(Obj_CType(tch->spec) != ctype_FLOWSPEC_Intserv0) {
		return ADMIT_FLOW_NOSVC;
	}
	z_specp = &(tch->spec->flow_body);

	/* Check against interface first */
	sp = (IS_serv_hdr_t *) &z_specp->spec_u;
	switch(sp->issh_service) {
	case GUARANTEED_SERV:
		if(!Allow_Guaranteed)
			return ADMIT_FLOW_NOSVC;
		/* fall into... */
	case CONTROLLED_LOAD_SERV:
		req_bandwidth = 
			z_specp->spec_u.CL_spec.CLspec_r * 8;
		if ((clinfo = ifinfo->resv_class) == NULL)
			return ADMIT_FLOW_SYSERR;
		acinfo = &(ifinfo->resv_class->acinfo);
		if ((acinfo->allocated + req_bandwidth) > acinfo->bandwidth)
			return ADMIT_FLOW_NOBW;
		break;

	default:
		return ADMIT_FLOW_NOSVC;
	}
	
	/* All checks passed */
	acinfo->allocated += req_bandwidth;
	acinfo->nflows++;
	return ADMIT_FLOW_OK;
}

static void
unadmit_send_flow(tc_t *tch)
{
	cbq_ifinfo_t		*ifinfo;
	cbq_acinfo_t		*acinfo;
	cbq_classinfo_t 	*clinfo;
	IS_specbody_t		*z_specp;
	IS_serv_hdr_t		*sp;
	long			req_bandwidth;

	assert(tch);

	if ((ifinfo = ifname2ifinfo(if_vec[tch->in_if].if_name)) == NULL)
		return;

	assert(Obj_CType(tch->spec) == ctype_FLOWSPEC_Intserv0);
	z_specp = &(tch->spec->flow_body);
	sp = (IS_serv_hdr_t *) &z_specp->spec_u;
	switch(sp->issh_service) {
	case GUARANTEED_SERV:
		if(!Allow_Guaranteed)
			return;
		/* fall into... */
	case CONTROLLED_LOAD_SERV:
		req_bandwidth = 
			z_specp->spec_u.CL_spec.CLspec_r * 8;
		clinfo = ifinfo->resv_class; 
		acinfo = &(ifinfo->resv_class->acinfo);
		acinfo->allocated -= req_bandwidth;
		acinfo->nflows--;
		break;

	default:
		return;
	}
}

#endif /* !CBQDAEMON */


