#pragma ident   "@(#)rsvp_cbq.h 1.22     97/06/19 SMI"
/*
 * User-visible data structures and definitions for CBQ code.
 */

#ifndef _rsvp_cbq_h
#define _rsvp_cbq_h

#include <netinet/altq_cbq.h>

/* interface info */
struct cbq_ifinfo {
	struct cbq_ifinfo	*next;
	char			*ifname;	/* interface name */
	long			bandwidth;	/* bandwidth in bps */
	double			nsPerByte;	/* bandwidth in ns per sec */
	int			is_wrr;		/* use weighted-round robin */
	int			is_efficient;	/* use work-conserving */
	int			use_cbq;	/* 0, if admission only */
	struct cbq_classinfo *resv_class;  /* meta class for reservation */
	struct cbq_classinfo *cllist;		/* class list */
	int			ifmtu;		/* mtu of the interface */
	int			enabled;	/* cbq on/off state */
	int			acc_enabled;	/* cbq account on/off state */
};
typedef struct cbq_ifinfo cbq_ifinfo_t;

typedef enum {
	CBQ_QOS_NONE,
	CBQ_QOS_GUARANTEED,
	CBQ_QOS_PREDICTIVE,
	CBQ_QOS_CNTR_DELAY,
	CBQ_QOS_CNTR_LOAD
}  cbq_tos_t;

/*
 * admission control info used for reservation
 * only used by a meta class for reservation
 */
struct cbq_acinfo {
	cbq_tos_t		type; 		/* e.g. CBQ_QOS_CNTR_LOAD */
	u_int			pri;		/* priority */
	long			bandwidth;	/* bandwidth in bps */
	long			allocated;	/* bandwidth already used */
	u_int			maxdelay;
	u_int			maxburst;
	u_int			minburst;
	u_int			av_pkt_size;	/* average packet size */
	u_int			max_pkt_size;	/* max packet size */
	u_int			nflows;		/* number of flows */
	struct cbq_flowinfo	*flow_list;	/* flow list */
};
typedef struct cbq_acinfo cbq_acinfo_t;

typedef enum {
	OL_NONE, OL_DROP, OL_DELAY
} cbq_ol_t;

/*
 * class info
 * more precisely, classinfo is for static classes.
 * dynamic classes for reservation are kept as flowinfo.
 */
struct cbq_classinfo {
	struct cbq_classinfo	*next;
	u_long			handle;		/* class handle */
	char			*clname;	/* class name */
	cbq_ifinfo_t		*ifinfo;	/* back pointer to ifinfo */
	struct cbq_classinfo	*parent_cl;
	struct cbq_classinfo	*borrow_cl;
	cbq_acinfo_t		acinfo;		/* admission control info */
};
typedef struct cbq_classinfo cbq_classinfo_t;

/*
 * filter info for flows
 * (no filter info is kept for static classes)
 */
struct cbq_fltrinfo {
	struct cbq_fltrinfo	*next;
	u_long			handle;		/* filter handle */
	cbq_filt_t		filter;		/* filter value */
	struct cbq_flowinfo	*flow;		/* back pointer to flowinfo */
};
typedef struct cbq_fltrinfo cbq_fltrinfo_t;

/* Per-service RSPEC definitions */
typedef struct {
    u_long g_rate;		/* GUARANTEED: flow rate parameter	*/
} cbq_g_rspec_t;

typedef struct {
    u_long p_dly;		/* PREDICTIVE: delay target parameter	*/
    				/* May also be used as priority..	*/
} cbq_p_rspec_t;

typedef union {
    cbq_g_rspec_t rs_g;	/* TOS_ST_GUARANTEED			*/
    cbq_p_rspec_t rs_p;	/* TOS_ST_PREDICTIVE			*/
} cbq_rspec_t;

/* Traffic spec (TSPEC) definitions */
typedef struct {
    u_long tbf_bkt;		/* tbf bucket size: bits		*/
    u_long tbf_rate;		/* tbf rate: bits/sec			*/
    u_long tbf_buf;		/* tbf reshaping buffer size (currently MBZ)*/
} cbq_tbf_t;

typedef struct {
    cbq_tbf_t ts_tb;
} cbq_tspec_t;

/*
 * dynamic class information
 */
struct cbq_flowinfo {
	struct cbq_flowinfo	*next;
	cbq_classinfo_t		*clinfo;	/* back pointer to classinfo
						   of parent */
/*XXX	cbq_tos_t		type;*/
	long			bandwidth; 
	u_long			handle;		/* class handle */
	cbq_fltrinfo_t		*fltr_list;	/* filters for this class */
};
typedef struct	cbq_flowinfo cbq_flowinfo_t;
typedef void	*cbq_handle_t;

extern int cbq_init(void);
extern int cbq_enableall(void);
extern int cbq_enable(char *);
extern int cbq_disableall(void);
extern int cbq_disable(char *);
extern int cbq_destroyall(void);
extern int cbq_add_if(char *, long, int, int, int);
extern int cbq_add_static_class(char *, char *, char *, char *,
	       u_int, int, u_int, u_int, u_int, u_int, u_int, u_int, 
	       cbq_tos_t, cbq_ol_t);
extern int cbq_delete_static_class(char *, char *);
extern int cbq_add_static_filter(char *, char *, cbq_filt_t *);
extern int cbq_acc_enable(char *);
extern int cbq_acc_disable(char *);
extern int cbq_is_enabled(char *);
extern int cbq_is_acc_enabled(char *);

extern void cbq_cmds_init(void);
extern int cbq_config(void);
extern int DoCommand(char *, FILE *);

#ifdef CBQDAEMON
#define IsDebug(type)  (l_debug >= LOG_DEBUG && (debug & (type)))
#define DEBUG_CBQ	0x40

#define ntoh16(x)	((u_int16_t)ntohs((u_int16_t)(x)))
#define ntoh32(x)	((u_int32_t)ntohl((u_int32_t)(x)))
#define hton16(x)	((u_int16_t)htons((u_int16_t)(x)))
#define hton32(x)	((u_int32_t)htonl((u_int32_t)(x)))

extern int	if_num;		/* number of phyints */
extern int	debug;		/* Debug output control bits */
extern int	l_debug;	/* Logging severity level */
extern int	Debug_mode;

extern char *cur_ifname(void);
extern void log(int, int, const char *, ...);

#endif /* CBQDAEMON */

#endif				/* _net_cbq_h */
