/*
 * Copyright (C) 1998
 *	Sony Computer Science Laboratory Inc.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY SONY CSL AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL SONY CSL OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: shadow2if.c,v 1.2 1998/04/03 07:34:23 kjc Exp kjc $
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/sockio.h>
#include <net/if.h>
#include <net/if_atm.h>

#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <err.h>

#if !defined(INET6) && !defined(HAVE_IF_NAMEINDEX)

struct if_nameindex {
  unsigned int   if_index;  /* 1, 2, ... */
  char          *if_name;   /* null terminated name: "le0", ... */
};

extern struct if_nameindex *if_nameindex(void);
extern void if_freenameindex(struct if_nameindex *);

#endif

struct sif_ent {
	struct sif_ent *next;
	char *sif_name;
	char *if_name;
};

static struct sif_ent *sif_list = NULL;

static void pvc_create_list(void);

char *shadow2if(char *if_name)
{
	struct sif_ent *s;
	static int first_time = 1;
	
	if (if_name[0] != 'p' || if_name[1] != 'v' || if_name[2] != 'c')
		return (if_name);

	if (first_time) {
		pvc_create_list();
		first_time = 0;
	}
	
	for (s = sif_list; s != NULL; s = s->next)
		if (strcmp(&if_name[3], &s->sif_name[3]) == 0)
			return (s->if_name);

	return (if_name);
}

static void
pvc_create_list(void)
{
	struct if_nameindex *ifn_list, *ifnp;
	struct ifreq ifr;
	struct sif_ent *s;
	int sock;

	if ((ifn_list = if_nameindex()) == NULL)
		err(1, "if_nameindex failed");

	if ((sock = socket(AF_INET, SOCK_DGRAM, 0)) < 0)
		err(1, "can't open socket");

	for (ifnp = ifn_list; ifnp->if_name != NULL; ifnp++) {
		if (strncmp(ifnp->if_name, "pvc", 3) != 0)
			continue;

		strcpy(ifr.ifr_name, ifnp->if_name);
		if (ioctl(sock, SIOCGPVCSIF, &ifr) < 0)
			err(1, "SIOCGPVCSIF");

		if ((s = malloc(sizeof(struct sif_ent))) == NULL)
			err(1, "malloc failed!");
		if ((s->sif_name = malloc(strlen(ifnp->if_name) + 1)) == NULL)
			err(1, "malloc failed!");
		strcpy(s->sif_name, ifnp->if_name);
		if ((s->if_name = malloc(strlen(ifr.ifr_name) + 1)) == NULL)
			err(1, "malloc failed!");
		strcpy(s->if_name, ifr.ifr_name);

		s->next = sif_list;
		sif_list = s;
	}

	close(sock);
	if_freenameindex(ifn_list);
}

