/* $Id: fifoqstat.c,v 0.2 1998/04/03 07:32:36 kjc Exp kjc $ */
/*
 * Copyright (C) 1997
 *	Sony Computer Science Laboratory Inc.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY SONY CSL AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL SONY CSL OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <limits.h>
#include <errno.h>
#include <err.h>

#include <sys/param.h>
#include <sys/ioctl.h>
#include <sys/fcntl.h>
#include <sys/time.h>

#include <sys/socket.h>
#include <net/if.h>

#include <netinet/altq_fifoq.h>

#define	FIFOQ_DEVICE	"/dev/fifoq"

int fd = -1;
char *if_name = NULL;
int interval = 5;

static void usage(void)
{
    fprintf(stderr, "usage: fifoqstat [-i interval] interface\n");
    exit(1);
}

static void
sigint_handler(int sig)
{
    fprintf(stderr, "Exiting on signal %d\n", sig);

    close(fd);

    exit(0);
}

int main(int argc, char **argv)
{
    struct fifoq_getstats get_stats;
    char *if_name;
    int ch;

    while ((ch = getopt(argc, argv, "i:")) != EOF) {
	switch (ch) {
	case 'i':
	    interval = atoi(optarg);
	    break;
	}
    }
    argc -= optind;
    argv += optind;
    if (argc > 0)
	if_name = argv[0];
    else {
	if (argc == 0)
	    fprintf(stderr, "missing interface name\n");
	usage();
    }

    signal(SIGINT, sigint_handler);
    signal(SIGTERM, sigint_handler);
    
    if ((fd = open(FIFOQ_DEVICE, O_RDONLY)) < 0)
	err(1, "fifoq open");

    if (strlen(if_name) > IFNAMSIZ)
	err(1, "interface name %s too long!\n", if_name);
    get_stats.iface.fifoq_ifname[IFNAMSIZ-1] = '\0';
    strncpy(get_stats.iface.fifoq_ifname, if_name, IFNAMSIZ-1);

    while (1) {
	
	if (ioctl(fd, FIFOQ_GETSTATS, &get_stats) < 0)
	    err(1, "ioctl FIFOQ_GETSTATS");

	printf(" q_len:%d, q_limit:%d\n",
	       get_stats.q_len, get_stats.q_limit);
	printf(" xmit: %qd pkts (%qd bytes) drop: %qd pkts (%qd bytes)\n",
	       get_stats.xmit_packets, get_stats.xmit_bytes,
	       get_stats.drop_packets, get_stats.drop_bytes);

	sleep(interval);
    }

    return (0);
}
